"""
Define constant mappings between bit-mask values and understandable quality
flags
"""
from functools import wraps


import pandas as pd
import numpy as np


# The quality_flag field in MySQL is currently limited to 1 << 15;
# fields beyond 1 << 15 will require a change in the MySQL datatype
# for the quality_flag column. The mapping from description to bitmask
# is versioned so that future addtions or removals are backwards compatible
# without rerunning the validation on all data.
# DO NOT MODIFY THE VALUES OF THE _BITMASK_DESCRIPTION_DICT instead,
# add a increment the key and add a new value tuple, i.e. add version 2 like
# 2: {'OK': 0, 'USER FLAGGED: 1 << 0, ...} . The VERSION
# IDENTIFIER 0 - 2 must remain in their current positions. Versions 7
# and up will require another identifier bit to be determined at that
# time.  The version identifier also serves to mark data in the
# database as validated. The tuples are (description, bit mask)
BITMASK_DESCRIPTION_DICT = {1: {
    # start with 1 to distinguish validated vs not in DB
    'OK': 0,
    'USER FLAGGED': 1 << 0,
    'VERSION IDENTIFIER 0': 1 << 1,
    'VERSION IDENTIFIER 1': 1 << 2,
    'VERSION IDENTIFIER 2': 1 << 3,
    'NIGHTTIME': 1 << 4,
    'CLEARSKY': 1 << 5,
    'SHADED': 1 << 6,
    'UNEVEN FREQUENCY': 1 << 7,
    'LIMITS EXCEEDED': 1 << 8,
    'CLEARSKY EXCEEDED': 1 << 9,
    'STALE VALUES': 1 << 10,
    'INTERPOLATED VALUES': 1 << 11,
    'CLIPPED VALUES': 1 << 12,
    'INCONSISTENT IRRADIANCE COMPONENTS': 1 << 13,
    'DAILY VALIDATION APPLIED': 1 << 14,
    'RESERVED 1': 1 << 15  # available for new flag
    }
}

# logical combinations of the masks defined above.
# add a version layer for compatibility if needed in the future.
# derived masks may reference masks defined in an earlier key.
DERIVED_MASKS = {
    'DAYTIME': (np.logical_not, 'NIGHTTIME'),
    'DAYTIME STALE VALUES': (np.logical_and, 'DAYTIME', 'STALE VALUES'),
    'DAYTIME INTERPOLATED VALUES': (
        np.logical_and, 'DAYTIME', 'INTERPOLATED VALUES'),
}

# should never change unless another VERSION IDENTIFIER is required
VERSION_MASK = 0b1110
LATEST_VERSION = max(BITMASK_DESCRIPTION_DICT.keys())
DESCRIPTION_MASK_MAPPING = BITMASK_DESCRIPTION_DICT[LATEST_VERSION]
LATEST_VERSION_FLAG = (
    LATEST_VERSION * DESCRIPTION_MASK_MAPPING['VERSION IDENTIFIER 0'])
DAILY_VALIDATION_FLAG = DESCRIPTION_MASK_MAPPING['DAILY VALIDATION APPLIED']


def convert_bool_flags_to_flag_mask(flags, flag_description, invert):
    if flags is None:
        return None
    if invert:
        bool_flags = ~(flags.astype(bool))
    else:
        bool_flags = flags.astype(bool)
    return ((bool_flags * DESCRIPTION_MASK_MAPPING[flag_description])
            | LATEST_VERSION_FLAG)


def mask_flags(flag_description, invert=True):
    """
    Decorator that will convert a boolean pandas object into an integer,
    bitmasked object when `_return_mask=True`. This decorator adds the
    `_return_mask` kwarg to the decorated function. Using this decorator
    to mask values ensures the description and decorated function are
    clearly linked.

    Parameters
    ----------
    flag_description : str
        Description of the flag to convert from a boolean to integer. Must be
        a key of the DESCRIPTION_MASK_MAPPING dict.
    invert : boolean
        Whether to invert the boolean object before conversion e.g. if
        flag_description = 'LIMITS EXCEEDED' and a True value indicates
        that a parameter is within the limits, invert=True is required
        for the proper mapping.

    Returns
    -------
    flags : pandas Object
        Returns the output of the decorated function (which must be a pandas
        Object) as the original output or an object of type int with value
        determined by the truthiness of the orignal output and flag_description
    """
    def decorator(f):
        @wraps(f)
        def wrapper(*args, **kwargs):
            return_mask = kwargs.pop('_return_mask', False)
            flags = f(*args, **kwargs)
            if return_mask:
                if isinstance(flags, tuple):
                    return tuple(convert_bool_flags_to_flag_mask(
                        f, flag_description, invert) for f in flags)
                else:
                    return convert_bool_flags_to_flag_mask(
                        flags, flag_description, invert)
            else:
                return flags
        return wrapper
    return decorator


def has_data_been_validated(flags):
    """Return True (or a boolean series) if flags has been validated"""
    return flags > 1


def get_version(flag):
    """Extract the version from flag"""
    # will be more complicated if another version identifier must be added
    return np.right_shift(flag & VERSION_MASK, 1)


def _flag_description_checks(flag_description):
    if isinstance(flag_description, str):
        return
    else:
        if len(flag_description) == 0:
            raise TypeError('flag_description must have len > 0')
        for k in iter(flag_description):
            if not isinstance(k, str):
                raise TypeError(
                    'Elements of flag_description must have type str')


def check_if_single_value_flagged(flag, flag_description,
                                  _perform_checks=True):
    """Check if the single integer flag has been flagged for flag_description

    Parameters
    ----------
    flag : integer
        Integer flag
    flag_description : string or iterable of strings
        Checks to compare againsts flag


    Returns
    -------
    Boolean
        Whether any of `flag_description` checks are represented by `flag`

    Raises
    ------
    ValueError
        If flag has not been validated
    TypeError
        If flag_description is not a string or iterable of strings
    KeyError
        If flag_description is not a possible check for the flag version
    """
    if _perform_checks:
        if not has_data_been_validated(flag):
            raise ValueError('Data has not been validated')
        _flag_description_checks(flag_description)
    mask_dict = BITMASK_DESCRIPTION_DICT[get_version(flag)]
    if isinstance(flag_description, str):
        mask = mask_dict[flag_description]
        ok_mask = mask == 0
    else:
        mask = 0
        ok_mask = False
        for k in flag_description:
            m = mask_dict[k]
            if m == 0:
                ok_mask = True
            mask |= m

    out = bool(flag & mask)
    if ok_mask:
        out |= which_data_is_ok(flag)
    return out


def which_data_is_ok(flags):
    """Return True for flags that have been validated and are OK"""
    return (flags & ~VERSION_MASK == 0) & has_data_been_validated(flags)


def _make_mask_series(version):
    descriptions = [k for k in BITMASK_DESCRIPTION_DICT[version].keys()
                    if not (k.startswith('VERSION') or
                            k.startswith('RESERVED') or k == 'OK')]
    masks = [BITMASK_DESCRIPTION_DICT[version][desc]
             for desc in descriptions]
    return pd.Series(masks, index=descriptions)


def check_for_all_descriptions(flag, _check_if_validated=True):
    """
    Return a boolean Series indicating the checks a flag represents
    """
    if _check_if_validated and not has_data_been_validated(flag):
        raise ValueError('Data has not been validated')
    version = get_version(flag)
    mask_series = _make_mask_series(version)
    out = (mask_series & flag).astype(bool)
    return out


def _convert_version_mask(ser):
    version = ser.name
    if version == 0:
        return pd.DataFrame({'NOT VALIDATED': [True] * len(ser.index)},
                            index=ser.index)
    mask_series = _make_mask_series(version)
    out = pd.DataFrame(
        np.bitwise_and(mask_series.values[None, :],
                       ser.values[:, None]),
        columns=mask_series.index,
        index=ser.index,
        dtype=bool)
    out['NOT VALIDATED'] = False
    return out


def _add_derived_masks(masks):
    """Copies input DataFrame and then adds new masks derived from
    input masks"""
    unvalidated = masks['NOT VALIDATED']
    if unvalidated.all():
        return masks
    out = masks.copy()[~unvalidated]
    for flag, operations in DERIVED_MASKS.items():
        func = operations[0]
        cols = operations[1:]
        args = [out[col] for col in cols]
        out[flag] = func(*args)
    return pd.concat([out, masks[unvalidated]], sort=False).fillna(False)


def convert_mask_into_dataframe(flag_series):
    """
    Convert `flag_series` into a boolean DataFrame indicating which checks
    the flags represent.

    Parameters
    ----------
    flag_series : pandas.Series
        Integer series of validated quality flags

    Returns
    -------
    pandas.DataFrame
       Columns are keys of BITMASK_DESCRIPTION_DICT and values are booleans
       indicating if the input flag corresponds to the given check. An
       additional column, NOT VALIDATED, indicates if the data has not
       been validated. Additional columns defined by DERIVED_MASKS are
       computed based on the results of the fundamental flags.
       Columns may vary depending the version of the quality
       flags in the series.
    """
    vers = get_version(flag_series)
    fundamental_masks = flag_series.groupby(vers, sort=False).apply(
        _convert_version_mask).fillna(False)
    out = _add_derived_masks(fundamental_masks)
    return out


def convert_flag_frame_to_strings(flag_frame, sep=', ', empty='OK'):
    """
    Convert the `flag_frame` output of :py:func:`~convert_mask_into_dataframe`
    into a pandas.Series of strings which are the active flag names separated
    by `sep`. Any row where all columns are false will have a value of `empty`.

    Parameters
    ----------
    flag_frame : pandas.DataFrame
        Boolean DataFrame with descriptive column names
    sep : str
        String to separate column names by
    empty : str
        String to replace rows where no columns are True

    Returns
    -------
    pandas.Series
        Of joined column names from `flag_frame` separated by `sep` if True.
        Has the same index as `flag_frame`.
    """
    return np.logical_and(flag_frame, flag_frame.columns + sep).replace(
        False, '').sum(axis=1).str.rstrip(sep).replace('', empty)


def check_if_series_flagged(flag_series, flag_description):
    """
    Check if `flag_series` has been flagged for the checks given by
    flag_description

    Parameters
    ----------
    flag_series : pandas.Series
        Series of integer quality flags
    flag_description : string or iterable of strings
        Checks to compare `flag_series` to. If this is an iterable, the result
        will be a boolean indicating if the flag represents *ANY* of the
        checks.

    Returns
    -------
    pandas.Series
        Boolean Series indicating if *ANY* of `flag_description` checks are
        represented by each flag

    Raises
    ------
    ValueError
        If any of `flag_series` has not been validated.
    TypeError
        If flag_description is not a string or iterable of strings
    KeyError
        If flag_description is not a possible check for the flag version
    """
    if not has_data_been_validated(flag_series).all():
        raise ValueError('Data has not been validated')
    _flag_description_checks(flag_description)
    return flag_series.apply(check_if_single_value_flagged,
                             flag_description=flag_description,
                             _perform_checks=False)



"""
Created on Wed Feb 27 15:12:14 2019

@author: cwhanse
"""

import numpy as np
import pandas as pd
from pandas.testing import assert_series_equal
from datetime import datetime
import pytz
import pytest
from solarforecastarbiter.validation import validator
import pvlib
from pvlib.location import Location


@pytest.fixture
def irradiance_QCRad():
    output = pd.DataFrame(
        columns=['ghi', 'dhi', 'dni', 'solar_zenith', 'dni_extra',
                 'ghi_limit_flag', 'dhi_limit_flag', 'dni_limit_flag',
                 'consistent_components', 'diffuse_ratio_limit'],
        data=np.array([[-100, 100, 100, 30, 1370, 0, 1, 1, 0, 0],
                       [100, -100, 100, 30, 1370, 1, 0, 1, 0, 0],
                       [100, 100, -100, 30, 1370, 1, 1, 0, 0, 1],
                       [1000, 100, 900, 0, 1370, 1, 1, 1, 1, 1],
                       [1000, 200, 800, 15, 1370, 1, 1, 1, 1, 1],
                       [1000, 200, 800, 60, 1370, 0, 1, 1, 0, 1],
                       [1000, 300, 850, 80, 1370, 0, 0, 1, 0, 1],
                       [1000, 500, 800, 90, 1370, 0, 0, 1, 0, 1],
                       [500, 100, 1100, 0, 1370, 1, 1, 1, 0, 1],
                       [1000, 300, 1200, 0, 1370, 1, 1, 1, 0, 1],
                       [500, 600, 100, 60, 1370, 1, 1, 1, 0, 0],
                       [500, 600, 400, 80, 1370, 0, 0, 1, 0, 0],
                       [500, 500, 300, 80, 1370, 0, 0, 1, 1, 1],
                       [0, 0, 0, 93, 1370, 1, 1, 1, 0, 0]]))
    dtypes = ['float64', 'float64', 'float64', 'float64', 'float64',
              'bool', 'bool', 'bool', 'bool', 'bool']
    for (col, typ) in zip(output.columns, dtypes):
        output[col] = output[col].astype(typ)
    return output


def test_check_ghi_limits_QCRad(irradiance_QCRad):
    expected = irradiance_QCRad
    ghi_out_expected = expected['ghi_limit_flag']
    ghi_out = validator.check_ghi_limits_QCRad(expected['ghi'],
                                               expected['solar_zenith'],
                                               expected['dni_extra'])
    assert_series_equal(ghi_out, ghi_out_expected)


def test_check_dhi_limits_QCRad(irradiance_QCRad):
    expected = irradiance_QCRad
    dhi_out_expected = expected['dhi_limit_flag']
    dhi_out = validator.check_dhi_limits_QCRad(expected['dhi'],
                                               expected['solar_zenith'],
                                               expected['dni_extra'])
    assert_series_equal(dhi_out, dhi_out_expected)


def test_check_dni_limits_QCRad(irradiance_QCRad):
    expected = irradiance_QCRad
    dni_out_expected = expected['dni_limit_flag']
    dni_out = validator.check_dni_limits_QCRad(expected['dni'],
                                               expected['solar_zenith'],
                                               expected['dni_extra'])
    assert_series_equal(dni_out, dni_out_expected)


def test_check_irradiance_limits_QCRad(irradiance_QCRad):
    expected = irradiance_QCRad
    ghi_out_expected = expected['ghi_limit_flag']
    ghi_out, dhi_out, dni_out = validator.check_irradiance_limits_QCRad(
        expected['solar_zenith'], expected['dni_extra'], ghi=expected['ghi'])
    assert_series_equal(ghi_out, ghi_out_expected)
    assert dhi_out is None
    assert dni_out is None

    dhi_out_expected = expected['dhi_limit_flag']
    ghi_out, dhi_out, dni_out = validator.check_irradiance_limits_QCRad(
        expected['solar_zenith'], expected['dni_extra'], ghi=expected['ghi'],
        dhi=expected['dhi'])
    assert_series_equal(dhi_out, dhi_out_expected)

    dni_out_expected = expected['dni_limit_flag']
    ghi_out, dhi_out, dni_out = validator.check_irradiance_limits_QCRad(
        expected['solar_zenith'], expected['dni_extra'],
        dni=expected['dni'])
    assert_series_equal(dni_out, dni_out_expected)


def test_check_irradiance_consistency_QCRad(irradiance_QCRad):
    expected = irradiance_QCRad
    cons_comp, diffuse = validator.check_irradiance_consistency_QCRad(
        expected['ghi'], expected['solar_zenith'], expected['dni_extra'],
        expected['dhi'], expected['dni'])
    assert_series_equal(cons_comp, expected['consistent_components'])
    assert_series_equal(diffuse, expected['diffuse_ratio_limit'])


@pytest.fixture
def weather():
    output = pd.DataFrame(columns=['air_temperature', 'wind_speed',
                                   'relative_humidity',
                                   'extreme_temp_flag', 'extreme_wind_flag',
                                   'extreme_rh_flag'],
                          data=np.array([[-40, -5, -5, 0, 0, 0],
                                         [10, 10, 50, 1, 1, 1],
                                         [140, 55, 105, 0, 0, 0]]))
    dtypes = ['float64', 'float64', 'float64', 'bool', 'bool', 'bool']
    for (col, typ) in zip(output.columns, dtypes):
        output[col] = output[col].astype(typ)
    return output


def test_check_temperature_limits(weather):
    expected = weather
    result_expected = expected['extreme_temp_flag']
    result = validator.check_temperature_limits(expected['air_temperature'])
    assert_series_equal(result, result_expected)


def test_check_wind_limits(weather):
    expected = weather
    result_expected = expected['extreme_wind_flag']
    result = validator.check_wind_limits(expected['wind_speed'])
    assert_series_equal(result, result_expected)


def test_check_rh_limits(weather):
    expected = weather
    data = expected['relative_humidity']
    result_expected = expected['extreme_rh_flag']
    result = validator.check_rh_limits(data)
    result.name = 'extreme_rh_flag'
    assert_series_equal(result, result_expected)


def test_check_ac_power_limits():
    index = pd.date_range(
        start='20200401 0700', freq='2h', periods=6, tz='UTC')
    power = pd.Series([0, -0.1, 0.1, 1, 1.1, -0.1], index=index)
    # not precise zenith - just for day/night flagging
    solar_zenith = pd.Series([180, 150, 120, 80, 50, 30], index=index)
    capacity = 1.
    expected = pd.Series([1, 0, 0, 1, 0, 0], index=index).astype(bool)
    out = validator.check_ac_power_limits(power, solar_zenith, capacity)
    assert_series_equal(out, expected)


def test_check_dc_power_limits():
    index = pd.date_range(
        start='20200401 0700', freq='2h', periods=6, tz='UTC')
    power = pd.Series([0, -0.1, 0.1, 1, 1.3, -0.1], index=index)
    # not precise zenith - just for day/night flagging
    solar_zenith = pd.Series([180, 150, 120, 80, 50, 30], index=index)
    capacity = 1.
    expected = pd.Series([1, 0, 0, 1, 0, 0], index=index).astype(bool)
    out = validator.check_dc_power_limits(power, solar_zenith, capacity)
    assert_series_equal(out, expected)


def test_check_limits():
    # testing with input type Series
    expected = pd.Series(data=[True, False])
    data = pd.Series(data=[3, 2])
    result = validator._check_limits(val=data, lb=2.5)
    assert_series_equal(expected, result)
    result = validator._check_limits(val=data, lb=3, lb_ge=True)
    assert_series_equal(expected, result)

    data = pd.Series(data=[3, 4])
    result = validator._check_limits(val=data, ub=3.5)
    assert_series_equal(expected, result)
    result = validator._check_limits(val=data, ub=3, ub_le=True)
    assert_series_equal(expected, result)

    result = validator._check_limits(val=data, lb=3, ub=4, lb_ge=True,
                                     ub_le=True)
    assert all(result)
    result = validator._check_limits(val=data, lb=3, ub=4)
    assert not any(result)

    with pytest.raises(ValueError):
        validator._check_limits(val=data)


@pytest.fixture
def location():
    return Location(latitude=35.05, longitude=-106.5, altitude=1619,
                    name="Albuquerque", tz="MST")


@pytest.fixture
def times():
    MST = pytz.timezone('MST')
    return pd.date_range(start=datetime(2018, 6, 15, 12, 0, 0, tzinfo=MST),
                         end=datetime(2018, 6, 15, 13, 0, 0, tzinfo=MST),
                         freq='10T')


def test_check_ghi_clearsky(mocker, location, times):
    clearsky = location.get_clearsky(times)
    # modify to create test conditions
    ghi = clearsky['ghi'].copy()
    ghi.iloc[0] *= 0.5
    ghi.iloc[-1] *= 2.0
    clear_times = np.tile(True, len(times))
    clear_times[-1] = False
    expected = pd.Series(index=times, data=clear_times)
    result = validator.check_ghi_clearsky(ghi, clearsky['ghi'])
    assert_series_equal(result, expected)


def test_check_poa_clearsky(mocker, times):
    dt = pd.date_range(start=datetime(2019, 6, 15, 12, 0, 0),
                       freq='15T', periods=5)
    poa_global = pd.Series(index=dt, data=[800, 1000, 1200, -200, np.nan])
    poa_clearsky = pd.Series(index=dt, data=1000)
    result = validator.check_poa_clearsky(poa_global, poa_clearsky)
    expected = pd.Series(index=dt, data=[True, True, False, True, False])
    assert_series_equal(result, expected)
    result = validator.check_poa_clearsky(poa_global, poa_clearsky, kt_max=1.2)
    expected = pd.Series(index=dt, data=[True, True, True, True, False])
    assert_series_equal(result, expected)


def test_check_irradiance_day_night():
    MST = pytz.timezone('MST')
    times = [datetime(2018, 6, 15, 12, 0, 0, tzinfo=MST),
             datetime(2018, 6, 15, 22, 0, 0, tzinfo=MST)]
    expected = pd.Series(data=[True, False], index=times)
    solar_zenith = pd.Series(data=[11.8, 114.3], index=times)
    result = validator.check_irradiance_day_night(solar_zenith)
    assert_series_equal(result, expected)


def test_check_timestamp_spacing(times):
    assert_series_equal(
        validator.check_timestamp_spacing(times, times.freq),
        pd.Series(True, index=times))

    assert_series_equal(
        validator.check_timestamp_spacing(times[[0]], times.freq),
        pd.Series(True, index=[times[0]]))

    assert_series_equal(
        validator.check_timestamp_spacing(times[[0, 2, 3]], times.freq),
        pd.Series([True, False, True], index=times[[0, 2, 3]]))

    assert_series_equal(
        validator.check_timestamp_spacing(times, '30min'),
        pd.Series([True] + [False] * (len(times) - 1), index=times))


def test_detect_stale_values():
    data = [1.0, 1.001, 1.001, 1.001, 1.001, 1.001001, 1.001, 1.001, 1.2, 1.3]
    x = pd.Series(data=data)
    res1 = validator.detect_stale_values(x, window=3)
    res2 = validator.detect_stale_values(x, rtol=1e-8, window=2)
    res3 = validator.detect_stale_values(x, window=7)
    res4 = validator.detect_stale_values(x, window=8)
    res5 = validator.detect_stale_values(x, rtol=1e-8, window=4)
    res6 = validator.detect_stale_values(x[1:], window=3)
    res7 = validator.detect_stale_values(x[1:8], window=3)
    assert_series_equal(res1, pd.Series([False, False, False, True, True, True,
                                         True, True, False, False]))
    assert_series_equal(res2, pd.Series([False, False, True, True, True, False,
                                         False, True, False, False]))
    assert_series_equal(res3, pd.Series([False, False, False, False, False,
                                         False, False, True, False, False]))
    assert not all(res4)
    assert_series_equal(res5, pd.Series([False, False, False, False, True,
                                         False, False, False, False, False]))
    assert_series_equal(res6, pd.Series(index=x[1:].index,
                                        data=[False, False, True, True, True,
                                              True, True, False, False]))
    assert_series_equal(res7, pd.Series(index=x[1:8].index,
                                        data=[False, False, True, True, True,
                                              True, True]))
    data = [0.0, 0.0, 0.0, -0.0, 0.00001, 0.000010001, -0.00000001]
    y = pd.Series(data=data)
    res = validator.detect_stale_values(y, window=3)
    assert_series_equal(res, pd.Series([False, False, True, True, False, False,
                                        False]))
    res = validator.detect_stale_values(y, window=3, atol=1e-3)
    assert_series_equal(res, pd.Series([False, False, True, True, True, True,
                                        True]))
    res = validator.detect_stale_values(y, window=3, atol=1e-5)
    assert_series_equal(res, pd.Series([False, False, True, True, True, False,
                                        False]))
    res = validator.detect_stale_values(y, window=3, atol=2e-5)
    assert_series_equal(res, pd.Series([False, False, True, True, True, True,
                                        True]))
    with pytest.raises(ValueError):
        validator.detect_stale_values(x, window=1)


def test_detect_interpolation():
    data = [1.0, 1.001, 1.002001, 1.003, 1.004, 1.001001, 1.001001, 1.001001,
            1.2, 1.3, 1.5, 1.4, 1.5, 1.6, 1.7, 1.8, 2.0]
    x = pd.Series(data=data)
    res1 = validator.detect_interpolation(x, window=3)
    assert_series_equal(res1, pd.Series([False, False, False, False, False,
                                         False, False, True, False, False,
                                         False, False, False, True, True, True,
                                         False]))
    res2 = validator.detect_interpolation(x, window=3, rtol=1e-2)
    assert_series_equal(res2, pd.Series([False, False, True, True, True,
                                         False, False, True, False, False,
                                         False, False, False, True, True, True,
                                         False]))
    res3 = validator.detect_interpolation(x, window=5)
    assert_series_equal(res3, pd.Series([False, False, False, False, False,
                                         False, False, False, False, False,
                                         False, False, False, False, False,
                                         True, False]))
    res4 = validator.detect_interpolation(x, window=3, atol=1e-2)
    assert_series_equal(res4, pd.Series([False, False, True, True, True,
                                         True, True, True, False, False,
                                         False, False, False, True, True, True,
                                         False]))
    data = [0.0, 0.0, 0.0, -0.0, 0.00001, 0.000010001, -0.00000001]
    y = pd.Series(data=data)
    res = validator.detect_interpolation(y, window=3, atol=1e-5)
    assert_series_equal(res, pd.Series([False, False, True, True, True, True,
                                        False]))
    res = validator.detect_stale_values(y, window=3, atol=1e-4)
    assert_series_equal(res, pd.Series([False, False, True, True, True, True,
                                        True]))
    with pytest.raises(ValueError):
        validator.detect_interpolation(x, window=2)


@pytest.fixture
def ghi_clearsky():
    MST = pytz.timezone('Etc/GMT+7')
    dt = pd.date_range(start=datetime(2019, 4, 3, 5, 0, 0, tzinfo=MST),
                       periods=60, freq='15T')
    loc = pvlib.location.Location(latitude=35, longitude=-110, tz=MST)
    cs = loc.get_clearsky(dt)
    return cs['ghi']


@pytest.fixture
def ghi_clipped(ghi_clearsky):
    ghi_clipped = ghi_clearsky.copy()
    ghi_clipped = np.minimum(ghi_clearsky, 800)
    ghi_clipped.iloc[12:17] = np.minimum(ghi_clearsky, 300)
    ghi_clipped.iloc[18:20] = np.minimum(ghi_clearsky, 300)
    ghi_clipped.iloc[26:28] *= 0.5
    ghi_clipped.iloc[36:] = np.minimum(ghi_clearsky, 400)
    return ghi_clipped


def test_detect_clipping(ghi_clipped):
    placeholder = pd.Series(index=ghi_clipped.index, data=False)
    expected = placeholder.copy()
    # for window=4 and fraction_in_window=0.75
    expected.iloc[3:6] = True
    expected.iloc[14:17] = True
    expected.iloc[18:20] = True
    expected.iloc[25] = True
    expected.iloc[30:36] = True
    expected.iloc[38:46] = True
    expected.iloc[56:60] = True
    flags = validator.detect_clipping(ghi_clipped, window=4,
                                      fraction_in_window=0.75, rtol=5e-3,
                                      levels=4)
    assert_series_equal(flags, expected)


def test_detect_clipping_some_nans(ghi_clipped):
    placeholder = pd.Series(index=ghi_clipped.index, data=False)
    expected = placeholder.copy()
    # for window=4 and fraction_in_window=0.75
    expected.iloc[3:6] = True
    expected.iloc[14:17] = True
    expected.iloc[18:20] = True
    expected.iloc[25] = True
    expected.iloc[30:36] = True
    expected.iloc[38:46] = True
    expected.iloc[56:60] = True
    inp = ghi_clipped.copy()
    inp.iloc[48] = np.nan
    flags = validator.detect_clipping(inp, window=4,
                                      fraction_in_window=0.75, rtol=5e-3,
                                      levels=4)
    assert_series_equal(flags, expected)


@pytest.mark.filterwarnings('ignore::RuntimeWarning')
def test_detect_clearsky_ghi(ghi_clearsky):
    flags = validator.detect_clearsky_ghi(ghi_clearsky, ghi_clearsky)
    # first 7 and last 6 values are judged not clear due to night (ghi=0)
    # and rapid change in ghi with sunrise and sunset
    assert all(flags[7:-6])
    assert not flags[0:7].any() and not flags[-6:].any()
    ghi_cloud = ghi_clearsky.copy()
    ghi_cloud[12:15] *= 0.5
    flags = validator.detect_clearsky_ghi(ghi_cloud, ghi_clearsky)
    assert all(flags[7:12]) and all(flags[15:-6])


def test_detect_clearsky_ghi_warn_interval_length(ghi_clearsky):
    with pytest.warns(RuntimeWarning):
        flags = validator.detect_clearsky_ghi(ghi_clearsky[::4],
                                              ghi_clearsky[::4])
    assert (flags == 0).all()


def test_detect_clearsky_ghi_warn_regular_interval(ghi_clearsky):
    with pytest.warns(RuntimeWarning):
        ser = ghi_clearsky[:-2].append(ghi_clearsky[-1:])
        flags = validator.detect_clearsky_ghi(ser, ser)
    assert (flags == 0).all()


def test_detect_clearsky_ghi_one_val(ghi_clearsky):
    ser = ghi_clearsky[:1]
    assert len(ser) == 1
    with pytest.warns(RuntimeWarning):
        flags = validator.detect_clearsky_ghi(ser, ser)
    assert (flags == 0).all()


@pytest.mark.parametrize('inp,expected', [
    (pd.Timedelta('15min'), 6),
    (pd.Timedelta('1h'), 3)
])
def test_stale_interpolated_window(inp, expected):
    assert validator.stale_interpolated_window(inp) == expected



import inspect
import os


import pandas as pd
import pytest
import json


from solarforecastarbiter.io.fetch import rtc

TEST_DATA_DIR = os.path.dirname(
    os.path.abspath(inspect.getfile(inspect.currentframe())))
SYSTEM_FILE = os.path.join(TEST_DATA_DIR, 'data', 'doe_rtc_system.json')
WEATHER_FILE = os.path.join(TEST_DATA_DIR, 'data', 'doe_rtc_weather.json')


start_date = pd.Timestamp('2019-01-22')
end_date = pd.Timestamp('2019-01-23')


def mocked_get(url):
    class MockedResponse:
        def __init__(self, json_data):
            self.json_data = json_data

        def json(self):
            return self.json_data

    if 'weather' in url:
        with open(WEATHER_FILE) as f:
            return MockedResponse(json.load(f))
    if 'system' in url:
        with open(SYSTEM_FILE) as f:
            return MockedResponse(json.load(f))


@pytest.mark.parametrize('loc,data_type,start,end,api_key,expected', [
    ('plant_a', 'weather', start_date, end_date, 'bogus_key',
     'https://pv-dashboard.sandia.gov/api/v1.0/location/plant_a/data/weather/start/2019-01-22/end/2019-01-23/key/bogus_key'), # NOQA
])
def test_request_doe_rtc(loc, data_type, start, end, api_key, expected,
                         mocker):
    mock_get = mocker.patch(
        'solarforecastarbiter.io.fetch.rtc.requests.get',
        side_effect=mocked_get)
    rtc.request_doe_rtc_data(loc, data_type, start, end, api_key)
    mock_get.assert_called_with(expected)


@pytest.mark.parametrize('location,api_key,start,end', [
    ('plant_a', 'bogus_key', start_date, end_date),
])
def test_fetch_doe_rtc(location, api_key, start, end, mocker):
    mocker.patch('solarforecastarbiter.io.fetch.rtc.requests.get',
                 side_effect=mocked_get)
    data = rtc.fetch_doe_rtc(location, api_key, start, end)
    assert data.index[0] == pd.Timestamp('2019-01-23T00:01:00', freq='T')
    assert data.index[-1] == pd.Timestamp('2019-01-23T23:59:00', freq='T')
    assert 'AmbientTemp_weather' in data.columns
    assert 'AmbientTemp_system' in data.columns
    assert 'TmStamp' not in data.columns



from dataclasses import fields, MISSING, dataclass
import json
from typing import Union


import numpy as np
import pandas as pd
import pytest


from solarforecastarbiter import datamodel


@pytest.fixture(params=['site', 'fixed', 'single', 'observation',
                        'forecast', 'forecastobservation',
                        'forecastaggregate',
                        'eventobservation', 'eventforecast',
                        'probabilisticforecastconstantvalue',
                        'probabilisticforecast', 'aggregate',
                        'aggregateforecast', 'aggregateprobforecast',
                        'aggregateobservation', 'report', 'quality_filter',
                        'timeofdayfilter', 'valuefilter', 'metricvalue',
                        'metricresult', 'validationresult',
                        'preprocessing_result', 'reportparameters',
                        'plotlyreportfigure', 'reportmessage',
                        'bokehreportfigure', 'constantcost',
                        'timeofdaycost', 'datetimecost', 'errorbandcost',
                        'cost'])
def pdid_params(request, many_sites, many_sites_text,
                single_observation, single_observation_text,
                single_site, single_forecast_text, single_forecast,
                single_event_observation_text, single_event_observation,
                single_event_forecast_text, single_event_forecast,
                prob_forecast_constant_value,
                prob_forecast_constant_value_text, prob_forecasts,
                prob_forecast_text, aggregate, aggregate_observations,
                aggregate_text, aggregate_forecast_text,
                aggregateforecast, aggregate_prob_forecast,
                aggregate_prob_forecast_text,
                agg_prob_forecast_constant_value,
                single_aggregate_observation,
                single_aggregate_observation_text, report_objects,
                report_dict, quality_filter, quality_filter_dict,
                timeofdayfilter, timeofdayfilter_dict, valuefilter,
                valuefilter_dict, metric_value_dict, metric_value,
                metric_result_dict, metric_result,
                validation_result_dict, validation_result,
                preprocessing_result_dict, preprocessing_result,
                plotly_report_figure_dict, plotly_report_figure,
                bokeh_report_figure_dict, bokeh_report_figure,
                report_message_dict, report_message,
                report_params_dict, report_params,
                constant_cost, timeofday_cost, datetime_cost,
                errorband_cost, banded_cost_params, cost_dicts):
    if request.param == 'site':
        return (many_sites[0], json.loads(many_sites_text)[0],
                datamodel.Site)
    elif request.param == 'fixed':
        return (many_sites[1].modeling_parameters,
                json.loads(many_sites_text)[1]['modeling_parameters'],
                datamodel.FixedTiltModelingParameters)
    elif request.param == 'single':
        return (many_sites[2].modeling_parameters,
                json.loads(many_sites_text)[2]['modeling_parameters'],
                datamodel.SingleAxisModelingParameters)
    elif request.param == 'observation':
        obs_dict = json.loads(single_observation_text)
        obs_dict['site'] = single_site
        return (single_observation, obs_dict,
                datamodel.Observation)
    elif request.param == 'forecast':
        fx_dict = json.loads(single_forecast_text)
        fx_dict['site'] = single_site
        return (single_forecast, fx_dict, datamodel.Forecast)
    elif request.param == 'eventobservation':
        obs_dict = json.loads(single_event_observation_text)
        obs_dict['site'] = single_site
        return (single_event_observation, obs_dict, datamodel.Observation)
    elif request.param == 'eventforecast':
        fx_dict = json.loads(single_event_forecast_text)
        fx_dict['site'] = single_site
        return (single_event_forecast, fx_dict, datamodel.EventForecast)
    elif request.param == 'probabilisticforecastconstantvalue':
        fx_dict = json.loads(prob_forecast_constant_value_text)
        fx_dict['site'] = single_site
        return (prob_forecast_constant_value, fx_dict,
                datamodel.ProbabilisticForecastConstantValue)
    elif request.param == 'probabilisticforecast':
        fx_dict = json.loads(prob_forecast_text)
        fx_dict['site'] = single_site
        fx_dict['constant_values'] = (prob_forecast_constant_value, )
        return (prob_forecasts, fx_dict, datamodel.ProbabilisticForecast)
    elif request.param == 'forecastobservation':
        fx_dict = json.loads(single_forecast_text)
        fx_dict['site'] = single_site
        obs_dict = json.loads(single_observation_text)
        obs_dict['site'] = single_site
        fxobs_dict = {'forecast': fx_dict, 'observation': obs_dict}
        fxobs = datamodel.ForecastObservation(
            single_forecast, single_observation)
        return (fxobs, fxobs_dict, datamodel.ForecastObservation)
    elif request.param == 'aggregate':
        agg_dict = json.loads(aggregate_text)
        agg_dict['observations'] = aggregate_observations
        return (aggregate, agg_dict, datamodel.Aggregate)
    elif request.param == 'aggregateforecast':
        aggfx_dict = json.loads(aggregate_forecast_text)
        aggfx_dict['aggregate'] = aggregate.to_dict()
        return (aggregateforecast, aggfx_dict, datamodel.Forecast)
    elif request.param == 'aggregateprobforecast':
        fx_dict = json.loads(aggregate_prob_forecast_text)
        fx_dict['aggregate'] = aggregate.to_dict()
        fx_dict['constant_values'] = (agg_prob_forecast_constant_value, )
        return (aggregate_prob_forecast, fx_dict,
                datamodel.ProbabilisticForecast)
    elif request.param == 'forecastaggregate':
        aggfx_dict = json.loads(aggregate_forecast_text)
        aggfx_dict['aggregate'] = aggregate.to_dict()
        agg_dict = json.loads(aggregate_text)
        agg_dict['observations'] = aggregate_observations
        fxobs_dict = {'forecast': aggfx_dict, 'aggregate': agg_dict}
        fxobs = datamodel.ForecastAggregate(aggregateforecast, aggregate)
        return (fxobs, fxobs_dict, datamodel.ForecastAggregate)
    elif request.param == 'aggregateobservation':
        aggobs_dict = json.loads(single_aggregate_observation_text)
        return (single_aggregate_observation, aggobs_dict,
                datamodel.AggregateObservation)
    elif request.param == 'report':
        report, *_ = report_objects
        return (report, report_dict.copy(), datamodel.Report)
    elif request.param == 'quality_filter':
        return (quality_filter, quality_filter_dict,
                datamodel.QualityFlagFilter)
    elif request.param == 'timeofdayfilter':
        return (timeofdayfilter, timeofdayfilter_dict,
                datamodel.TimeOfDayFilter)
    elif request.param == 'valuefilter':
        return (valuefilter, valuefilter_dict,
                datamodel.ValueFilter)
    elif request.param == 'metricvalue':
        return (metric_value, metric_value_dict, datamodel.MetricValue)
    elif request.param == 'metricresult':
        return (metric_result, metric_result_dict, datamodel.MetricResult)
    elif request.param == 'validationresult':
        return (validation_result, validation_result_dict,
                datamodel.ValidationResult)
    elif request.param == 'preprocessing_result':
        return (preprocessing_result, preprocessing_result_dict,
                datamodel.PreprocessingResult)
    elif request.param == 'reportparameters':
        return (report_params, report_params_dict,
                datamodel.ReportParameters)
    elif request.param == 'plotlyreportfigure':
        return (plotly_report_figure, plotly_report_figure_dict,
                datamodel.PlotlyReportFigure)
    elif request.param == 'bokehreportfigure':
        return (bokeh_report_figure, bokeh_report_figure_dict,
                datamodel.BokehReportFigure)
    elif request.param == 'reportmessage':
        return (report_message, report_message_dict, datamodel.ReportMessage)
    elif request.param == 'constantcost':
        return (constant_cost, cost_dicts['constant'], datamodel.ConstantCost)
    elif request.param == 'timeofdaycost':
        return (timeofday_cost, cost_dicts['timeofday'],
                datamodel.TimeOfDayCost)
    elif request.param == 'datetimecost':
        return (datetime_cost, cost_dicts['datetime'], datamodel.DatetimeCost)
    elif request.param == 'errorbandcost':
        return (errorband_cost, cost_dicts['errorband'],
                datamodel.ErrorBandCost)
    elif request.param == 'cost':
        return (banded_cost_params, cost_dicts['fullcost'], datamodel.Cost)


@pytest.mark.parametrize('extra', [
    {},
    {'extra': 'thing'},
])
def test_from_dict_into_datamodel(extra, pdid_params):
    expected, obj_dict, model = pdid_params
    obj_dict.update(extra)
    out = model.from_dict(obj_dict)
    assert out == expected


@pytest.fixture(params=[0, 1, 2])
def basefilter_params(
        request, valuefilter, valuefilter_dict, quality_filter,
        quality_filter_dict, timeofdayfilter, timeofdayfilter_dict):
    parameters = [
        (valuefilter, valuefilter_dict),
        (quality_filter, quality_filter_dict),
        (timeofdayfilter, timeofdayfilter_dict)]
    return parameters[request.param]


def test_base_filter_from_dict_into_datamodel(basefilter_params):
    expected, obj_dict = basefilter_params
    out = datamodel.BaseFilter.from_dict(obj_dict, raise_on_extra=True)
    assert out == expected


def test_filter_to_dict(basefilter_params):
    expected, obj_dict = basefilter_params
    out = expected.to_dict()
    assert out == obj_dict


def test_from_dict_into_datamodel_missing_field(pdid_params):
    _, obj_dict, model = pdid_params
    field_to_remove = None
    for field in fields(model):
        if field.default is MISSING and field.default_factory is MISSING:
            field_to_remove = field.name
            break
    if field_to_remove is not None:
        del obj_dict[field_to_remove]
        with pytest.raises(KeyError):
            model.from_dict(obj_dict)


def test_from_dict_into_datamodel_no_extra(pdid_params):
    expected, obj_dict, model = pdid_params
    obj_dict.pop('extra_parameters', '')
    out = model.from_dict(obj_dict)
    for field in fields(model):
        if field.name == 'extra_parameters':
            continue
        expected_value = getattr(expected, field.name)
        out_value = getattr(out, field.name)
        # normalization can be nan. isinstance ensures that isnan doesn't
        # try to coerce random strings or objects to a float and then error
        if isinstance(expected_value, float) and np.isnan(expected_value):
            assert np.isnan(out_value)
        else:
            assert out_value == expected_value


def test_from_dict_no_extra(pdid_params):
    expected, obj_dict, model = pdid_params
    names = [f.name for f in fields(model)]
    for key in list(obj_dict.keys()):
        if key not in names:
            del obj_dict[key]
    assert model.from_dict(obj_dict, raise_on_extra=True) == expected


def test_from_dict_extra_params_raise(pdid_params):
    _, obj_dict, model = pdid_params
    obj_dict['superextra'] = 'thing'
    with pytest.raises(KeyError):
        model.from_dict(obj_dict, raise_on_extra=True)


@pytest.mark.parametrize('site_num', [1, 2])
def test_from_dict_invalid_tracking_type(site_num, many_sites_text):
    model = datamodel.SolarPowerPlant
    obj_dict = json.loads(many_sites_text)[site_num]
    obj_dict['modeling_parameters']['tracking_type'] = 'invalid'
    with pytest.raises(ValueError):
        model.from_dict(obj_dict)


@pytest.mark.parametrize('model', [datamodel.Observation, datamodel.Forecast])
def test_from_dict_invalid_timedelta(model, many_observations_text,
                                     many_forecasts_text):
    if model == datamodel.Observation:
        obj_dict = json.loads(many_observations_text)[0]
    else:
        obj_dict = json.loads(many_forecasts_text)[0]
    obj_dict['interval_length'] = 'blah'
    with pytest.raises(ValueError):
        model.from_dict(obj_dict)


def test_from_dict_invalid_time_format(many_forecasts_text):
    obj_dict = json.loads(many_forecasts_text)[0]
    obj_dict['issue_time_of_day'] = '0000'
    with pytest.raises(ValueError):
        datamodel.Forecast.from_dict(obj_dict)


def test_from_dict_invalid_constant_values(prob_forecast_text, single_site):
    fx_dict = json.loads(prob_forecast_text)
    fx_dict['site'] = single_site
    fx_dict['constant_values'] = ('not a tuple of cv', 1)
    with pytest.raises(TypeError):
        datamodel.ProbabilisticForecast.from_dict(fx_dict)


def test_from_dict_invalid_axis(prob_forecast_text, single_site,
                                prob_forecast_constant_value):
    fx_dict = json.loads(prob_forecast_text)
    fx_dict['site'] = single_site
    fx_dict['constant_values'] = (prob_forecast_constant_value,)
    fx_dict['axis'] = 'z'
    with pytest.raises(ValueError):
        datamodel.ProbabilisticForecast.from_dict(fx_dict)


def test_from_dict_inconsistent_axis(prob_forecast_text, single_site,
                                     prob_forecast_constant_value_text,
                                     prob_forecast_constant_value):
    cv_dict = json.loads(prob_forecast_constant_value_text)
    cv_dict['site'] = single_site
    fx_dict = json.loads(prob_forecast_text)
    fx_dict['site'] = single_site
    fx_dict['constant_values'] = (prob_forecast_constant_value, cv_dict)
    fx_dict['axis'] = 'x'
    # check multiple constant values
    datamodel.ProbabilisticForecast.from_dict(fx_dict)
    cv_dict['axis'] = 'y'
    with pytest.raises(ValueError):
        datamodel.ProbabilisticForecast.from_dict(fx_dict)


def test_dict_roundtrip(pdid_params):
    expected, _, model = pdid_params
    dict_ = expected.to_dict()
    out = model.from_dict(dict_)
    assert out == expected


def test_to_dict_recurse(single_forecast):
    @dataclass
    class Special(datamodel.BaseModel):
        fx: datamodel.Forecast

    spec = Special(single_forecast)
    d_ = spec.to_dict()
    assert isinstance(d_['fx']['run_length'], float)
    assert isinstance(d_['fx']['issue_time_of_day'], str)
    assert isinstance(d_['fx'], dict)
    assert isinstance(d_['fx']['site'], dict)


def test_replace(single_forecast):
    run_length = single_forecast.run_length
    new_fx = single_forecast.replace(run_length=pd.Timedelta('3d'))
    assert new_fx.run_length != run_length
    assert new_fx != single_forecast
    assert new_fx.name == single_forecast.name


def test_invalid_variable(single_site):
    with pytest.raises(ValueError):
        datamodel.Observation(
            name='test', variable='noway',
            interval_value_type='mean',
            interval_length=pd.Timedelta('1min'),
            interval_label='beginning',
            site=single_site,
            uncertainty=0.1,
        )


@pytest.fixture(params=[0, 1, 2])
def _sites(many_sites_text, many_sites, request):
    models = [datamodel.Site, datamodel.SolarPowerPlant,
              datamodel.SolarPowerPlant]
    param = request.param
    site_dict_list = json.loads(many_sites_text)
    return site_dict_list[param], many_sites[param], models[param]


def test_process_site_dict(_sites):
    site_dict, expected, model = _sites
    out = model.from_dict(site_dict)
    assert out == expected


def test_process_nested_objects(single_observation_text_with_site_text,
                                single_site, single_observation):
    obs_dict = json.loads(single_observation_text_with_site_text)
    obs = datamodel.Observation.from_dict(obs_dict)
    assert obs == single_observation
    assert obs.site == single_site
    assert obs.site == single_observation.site


@pytest.mark.parametrize('key,val', [
    ('interval_length', pd.Timedelta('2h')),
    ('interval_value_type', 'interval_max'),
    ('variable', 'ghi')
])
def test_aggregate_invalid(single_observation, key, val):
    obsd = single_observation.to_dict()
    obsd[key] = val
    obs = datamodel.Observation.from_dict(obsd)
    aggobs = datamodel.AggregateObservation(
        obs, pd.Timestamp.utcnow())
    with pytest.raises(ValueError):
        datamodel.Aggregate(
            'test', 'testd', 'dni', 'mean', pd.Timedelta('1h'),
            'ending', 'America/Denver',
            observations=(aggobs,)
        )


def test_forecast_invalid(single_forecast, single_site, aggregate):
    with pytest.raises(KeyError):
        single_forecast.replace(site=None, aggregate=None)
    with pytest.raises(KeyError):
        single_forecast.replace(site=single_site, aggregate=aggregate)


def test_probabilistic_forecast_float_constant_values(prob_forecasts):
    out = datamodel.ProbabilisticForecast(
        name=prob_forecasts.name,
        issue_time_of_day=prob_forecasts.issue_time_of_day,
        lead_time_to_start=prob_forecasts.lead_time_to_start,
        interval_length=prob_forecasts.interval_length,
        run_length=prob_forecasts.run_length,
        interval_label=prob_forecasts.interval_label,
        interval_value_type=prob_forecasts.interval_value_type,
        variable=prob_forecasts.variable,
        site=prob_forecasts.site,
        forecast_id=prob_forecasts.forecast_id,
        axis=prob_forecasts.axis,
        extra_parameters=prob_forecasts.extra_parameters,
        provider=prob_forecasts.provider,
        constant_values=tuple(cv.constant_value
                              for cv in prob_forecasts.constant_values),
    )
    object.__setattr__(prob_forecasts.constant_values[0], 'forecast_id', '')
    assert isinstance(out.constant_values[0],
                      datamodel.ProbabilisticForecastConstantValue)
    assert out == prob_forecasts


def test_probabilistic_forecast_float_constant_values_from_dict(
        prob_forecast_text, single_site, prob_forecast_constant_value,
        prob_forecasts):
    fx_dict = json.loads(prob_forecast_text)
    fx_dict['site'] = single_site
    fx_dict['constant_values'] = (
        prob_forecast_constant_value.constant_value, )
    out = datamodel.ProbabilisticForecast.from_dict(fx_dict)
    object.__setattr__(prob_forecasts.constant_values[0], 'forecast_id', '')
    assert out == prob_forecasts


def test_probabilistic_forecast_invalid_constant_value(prob_forecasts):
    with pytest.raises(TypeError):
        prob_forecasts.replace(constant_values=(1, 'a'))


def test__single_field_processing_union():
    @dataclass
    class Model(datamodel.BaseModel):
        myfield: Union[int, pd.Timestamp, None]

    out = datamodel._single_field_processing(Model, fields(Model)[0],
                                             '20190101')
    assert out == pd.Timestamp('20190101')

    out = datamodel._single_field_processing(Model, fields(Model)[0], None)
    assert out is None

    out = datamodel._single_field_processing(Model, fields(Model)[0], 10)
    assert out == 10

    with pytest.raises(TypeError):
        datamodel._single_field_processing(Model, fields(Model)[0], 'bad')


def test_forecast_from_union(single_forecast, single_forecast_text, site_text):
    @dataclass
    class Model(datamodel.BaseModel):
        myfield: Union[datamodel.Observation, datamodel.Forecast]

    fxdict = json.loads(single_forecast_text)
    fxdict['site'] = json.loads(site_text)
    out = Model.from_dict({'myfield': fxdict})
    assert out.myfield == single_forecast


def test___check_categories__():
    datamodel.__check_categories__(['total', 'weekday'])
    with pytest.raises(ValueError):
        datamodel.__check_categories__(['bad', 'very bad'])


@pytest.mark.parametrize('metrics', [
    (['rmse']),
    (list(datamodel.ALLOWED_DETERMINISTIC_METRICS.keys())),
    pytest.param(
        ["bss"],
        marks=pytest.mark.xfail(raises=ValueError, strict=True)
    ),
    pytest.param(
        list(datamodel.ALLOWED_PROBABILISTIC_METRICS.keys()),
        marks=pytest.mark.xfail(raises=ValueError, strict=True)
    ),
])
def test___check_metrics__(metrics, single_forecast):
    datamodel.__check_metrics__(single_forecast, metrics)


@pytest.mark.parametrize('metrics', [
    (["far"]),
    (list(datamodel.ALLOWED_EVENT_METRICS.keys())),
    pytest.param(
        ["rmse"],
        marks=pytest.mark.xfail(raises=ValueError, strict=True)
    ),
])
def test___check_metrics___event(single_event_forecast, metrics):
    datamodel.__check_metrics__(single_event_forecast, metrics)


@pytest.mark.parametrize('metrics', [
    (['crps']),
    (list(datamodel.ALLOWED_PROBABILISTIC_METRICS.keys())),
    pytest.param(
        ('rmse'),
        marks=pytest.mark.xfail(raises=ValueError, strict=True)
    ),
    pytest.param(
        list(datamodel.ALLOWED_DETERMINISTIC_METRICS.keys()),
        marks=pytest.mark.xfail(raises=ValueError, strict=True)
    ),
])
def test___check_metrics__probabilistic(metrics, prob_forecast_constant_value):
    datamodel.__check_metrics__(prob_forecast_constant_value, metrics)


@pytest.fixture
def objects_from_attrs(mocker):
    """Takes a list of lists with tuples of (attr_name, value)
       and creates a listed of Mock objects with their attributes
       set to those values
    """
    def fn(attr_list):
        things_with_attrs = []
        for attr_tuples in attr_list:
            with_attr = mocker.Mock()
            for (attr_name, attr_value) in attr_tuples:
                setattr(with_attr, attr_name, attr_value)
            things_with_attrs.append(with_attr)
        return things_with_attrs
    return fn


def test___check_units__(mocker, objects_from_attrs):
    things_with_units = objects_from_attrs(
        [[('units', u)] for u in ['W/M^2', 'W/M^2', 'W/M^2', 'W/M^2', 'W/M^2']]
    )
    datamodel.__check_units__(*things_with_units)


def test___check_units__no_args():
    datamodel.__check_units__()


def test___check_units___error(mocker, objects_from_attrs):
    things_with_units = objects_from_attrs(
        [[('units', u)] for u in ['W/M^2', 'different', 'W/M^2', 'W/M^2']]
    )
    with pytest.raises(ValueError):
        datamodel.__check_units__(*things_with_units)


@pytest.mark.parametrize('fx_int, fx_label, obs_int, obs_label', [
    (15, 'instant', 15, 'instant'), (1, 'ending', 1, 'ending'),
    (5, 'beginning', 5, 'beginning'), (15, 'beginning', 5, 'instant'),
])
def test___check_interval_compatibility__(
        fx_int, fx_label, obs_int, obs_label, objects_from_attrs):
    attrs = [(('interval_length', fx_int), ('interval_label', fx_label)),
             (('interval_length', obs_int), ('interval_label', obs_label))]
    forecast_and_observation = objects_from_attrs(attrs)
    datamodel.__check_interval_compatibility__(*forecast_and_observation)


def test___check_interval_compatibility__bad_labels(objects_from_attrs):
    attrs = [(('interval_length', 5), ('interval_label', 'instant')),
             (('interval_length', 5), ('interval_label', 'beginning'))]
    forecast_and_observation = objects_from_attrs(attrs)
    with pytest.raises(ValueError):
        datamodel.__check_interval_compatibility__(*forecast_and_observation)


def test___check_interval_compatibility__bad_length(objects_from_attrs):
    attrs = [(('interval_length', 5), ('interval_label', 'instant')),
             (('interval_length', 15), ('interval_label', 'instant'))]
    forecast_and_observation = objects_from_attrs(attrs)
    with pytest.raises(ValueError):
        datamodel.__check_interval_compatibility__(*forecast_and_observation)


def test_base_filter_from_dict_error():
    with pytest.raises(NotImplementedError):
        datamodel.BaseFilter.from_dict({'spicyness': 'very'})


def test_quality_flag_filter_post_init_error():
    with pytest.raises(ValueError):
        datamodel.QualityFlagFilter.from_dict(
            {'quality_flags': ['spicy_filter']})


def test_metric_result_post_init_error(metric_result_dict):
    metric_result_dict['aggregate_id'] = None
    metric_result_dict['observation_id'] = None
    with pytest.raises(ValueError):
        datamodel.MetricResult.from_dict(metric_result_dict)


# These 'sfp' tests check that behavior found in the _special_field_processing
# functions is maintained after removing the code.
@pytest.mark.parametrize('sitef', [
    lambda x: x,
    lambda x: x.to_dict(),
    pytest.param(lambda x: 'other', marks=pytest.mark.xfail(strict=True))
])
def test_forecast_sfp_site_dict(single_forecast_text, single_site, sitef):
    forecast_dict = json.loads(single_forecast_text)
    forecast_dict['site'] = sitef(single_site)
    forecast = datamodel.Forecast.from_dict(forecast_dict)
    assert isinstance(forecast.site, datamodel.Site)


@pytest.mark.parametrize('aggf', [
    lambda x: x,
    lambda x: x.to_dict(),
    pytest.param(lambda x: 'other', marks=pytest.mark.xfail(strict=True))
])
def test_forecast_sfp_aggregate_dict(single_forecast_text, aggregate, aggf):
    forecast_dict = json.loads(single_forecast_text)
    forecast_dict['aggregate'] = aggf(aggregate)
    forecast = datamodel.Forecast.from_dict(forecast_dict)
    assert isinstance(forecast.aggregate, datamodel.Aggregate)


@pytest.mark.parametrize('key,value,expected', [
    ('effective_until', None, None),
    ('effective_until', '2020-01-01T00:00Z',
     pd.Timestamp('2020-01-01T00:00Z')),
    ('observation_deleted_at', None, None),
    ('observation_deleted_at', '2020-01-01T00:00Z',
     pd.Timestamp('2020-01-01T00:00Z'))
])
def test_aggregate_observation_sfp(
        aggregate_observations, key, value, expected):
    aggobs = aggregate_observations[0]
    aggobs_dict = aggobs.to_dict()
    aggobs_dict[key] = value
    aggobs_from_dict = datamodel.AggregateObservation.from_dict(aggobs_dict)
    assert getattr(aggobs_from_dict, key) == expected


@pytest.mark.parametrize('key,value,expected', [
    ('effective_until', 'bad', None),
    ('observation_deleted_at', 'bad', None),
])
def test_aggregate_observation_sfp_invalid(
        aggregate_observations, key, value, expected):
    aggobs = aggregate_observations[0]
    aggobs_dict = aggobs.to_dict()
    aggobs_dict[key] = value
    with pytest.raises(TypeError):
        datamodel.AggregateObservation.from_dict(aggobs_dict)


@pytest.mark.parametrize('spec', [
    ('\"<script>alert("hello");</script>\"'),
    ('<script>alert("hello");</script>'),
    ('bad'),
    ('\"bad\"'),
    ('{"key": "value": "value"}'),
    ('{"key": "value"});</script><script>parseInt(1'),  # noqa
    ('{});</script><a onclick=alert("hello")></a><script>eval('),
])
def test_load_report_figure_invalid_spec(plotly_report_figure_dict, spec):
    json_dict = plotly_report_figure_dict.copy()
    json_dict.update(spec=spec)
    with pytest.raises(ValueError):
        datamodel.PlotlyReportFigure.from_dict(json_dict)


@pytest.mark.parametrize('spec', [
    ('\"<script>alert("hello");</script>\"'),
    ('<script>alert("hello");</script>'),
    ('bad'),
    ('\"bad\"'),
    ('{"key": "value": "value"}'),
    ('{"key": "value"});</script><script>parseInt(1'),  # noqa
    ('{});</script><a onclick=alert("hello")></a><script>eval('),
])
def test_update_report_figure_invalid_spec(plotly_report_figure_dict, spec):
    fig = datamodel.PlotlyReportFigure.from_dict(plotly_report_figure_dict)
    with pytest.raises(ValueError):
        fig.replace(spec=spec)


def test_invalid_empty_report_figure():
    with pytest.raises(ValueError):
        datamodel.ReportFigure()


@pytest.mark.parametrize('plotly_version,bokeh_version,script', [
    ('1.0.0', '1.0.0', '<script></script>'),
    (None, None, '<script></script>'),
    (None, '1.0.0', None),
])
def test_raw_report_plots_invalid_versions(
        plotly_report_figure, plotly_version, bokeh_version, script):
    with pytest.raises(KeyError):
        datamodel.RawReportPlots(
            figures=(plotly_report_figure,),
            plotly_version=plotly_version,
            bokeh_version=bokeh_version,
            script=script
        )


@pytest.mark.parametrize('plotly_version,bokeh_version,script', [
    ('1.0.0', None, None),
    (None, '1.0.0', '<script></script>')
])
def test_raw_report_plots(
        plotly_report_figure, plotly_version, bokeh_version, script):
    dm_obj = datamodel.RawReportPlots(
        figures=(plotly_report_figure,),
        plotly_version=plotly_version,
        bokeh_version=bokeh_version,
        script=script
    )
    assert dm_obj.figures == (plotly_report_figure,)
    assert dm_obj.plotly_version == plotly_version
    assert dm_obj.bokeh_version == bokeh_version
    assert dm_obj.script == script


def test_bokeh_report_figure(bokeh_report_figure_dict):
    dm_obj = datamodel.BokehReportFigure.from_dict(
        bokeh_report_figure_dict)
    for k, v in bokeh_report_figure_dict.items():
        assert getattr(dm_obj, k) == v


def test_plotly_report_figure(plotly_report_figure_dict):
    dm_obj = datamodel.PlotlyReportFigure.from_dict(
        plotly_report_figure_dict)
    for k, v in plotly_report_figure_dict.items():
        assert getattr(dm_obj, k) == v


def test_bokeh_report_figure_plotly_dict(plotly_report_figure_dict):
    with pytest.raises(KeyError):
        datamodel.BokehReportFigure.from_dict(plotly_report_figure_dict)


def test_plotly_report_figure_bokeh_dict(bokeh_report_figure_dict):
    with pytest.raises(KeyError):
        datamodel.PlotlyReportFigure.from_dict(bokeh_report_figure_dict)


def test_report_figure_from_dict_with_bokeh(bokeh_report_figure_dict):
    dm_obj = datamodel.ReportFigure.from_dict(bokeh_report_figure_dict)
    for key, value in bokeh_report_figure_dict.items():
        assert getattr(dm_obj, key) == value
    assert isinstance(dm_obj, datamodel.BokehReportFigure)


def test_report_figure_from_dict_with_plotly(plotly_report_figure_dict):
    dm_obj = datamodel.ReportFigure.from_dict(plotly_report_figure_dict)
    for key, value in plotly_report_figure_dict.items():
        assert getattr(dm_obj, key) == value
    assert isinstance(dm_obj, datamodel.PlotlyReportFigure)


@pytest.mark.parametrize('fig_dict', [
    {"no_class": "bad"},
    {"non-implemented": "d3"},
])
def test_report_figure_from_dict_invalid(fig_dict):
    with pytest.raises(NotImplementedError):
        datamodel.ReportFigure.from_dict(fig_dict)


def test_ForecastObservation_normalization_ac(single_forecast_ac_observation):
    assert single_forecast_ac_observation.normalization == 0.003


def test_ForecastObservation_normalization_dc(single_forecast_dc_observation):
    assert single_forecast_dc_observation.normalization == 0.0035


def test_ForecastObservation_normalization_wind_speed(
        single_forecast_wind_speed_observation):
    assert np.isnan(single_forecast_wind_speed_observation.normalization)


@pytest.mark.parametrize('uncertainty,expected', [
    (5, 5.),
    ('5', 5.),
    ('observation_uncertainty', 0.1),
    (None, None)
])
def test_ForecastObservation_uncertainty(
        single_forecast, single_observation, uncertainty, expected):
    fxobs = datamodel.ForecastObservation(
        single_forecast, single_observation, uncertainty=uncertainty)
    assert fxobs.uncertainty == expected


def test_ForecastObservation_uncertainty_invalid(
        single_forecast, single_observation):
    with pytest.raises(ValueError):
        datamodel.ForecastObservation(
            single_forecast, single_observation, uncertainty='nope')


@pytest.mark.parametrize('uncertainty,expected', [
    (5, 5.),
    ('5', 5.),
    (None, None)
])
def test_ForecastAggregate_uncertainty(
        aggregateforecast, aggregate, uncertainty, expected):
    fxobs = datamodel.ForecastAggregate(
        aggregateforecast, aggregate, uncertainty=uncertainty)
    assert fxobs.uncertainty == expected


def test_ForecastAggregate_uncertainty_invalid(
        aggregateforecast, aggregate):
    with pytest.raises(ValueError):
        datamodel.ForecastAggregate(
            aggregateforecast, aggregate, uncertainty='anystring')


def test_ForecastObservation_reffx(single_forecast_observation_reffx):
    assert isinstance(single_forecast_observation_reffx.reference_forecast,
                      datamodel.Forecast)


def test_ForecastAggregate_reffx(aggregateforecast, aggregate):
    fxobs = datamodel.ForecastAggregate(
        aggregateforecast, aggregate, reference_forecast=aggregateforecast)
    assert isinstance(fxobs.reference_forecast, datamodel.Forecast)


@pytest.mark.parametrize("interval_label,variable", [
    ("event", "event"),
    pytest.param("event", "ghi",
                 marks=pytest.mark.xfail(raises=ValueError, strict=True)),
    pytest.param("beginning", "event",
                 marks=pytest.mark.xfail(raises=ValueError, strict=True)),
    pytest.param("beginning", "dni",
                 marks=pytest.mark.xfail(raises=ValueError, strict=True)),
])
def test_EventForecast_parameters(site_metadata, interval_label, variable):
    fx = datamodel.EventForecast(
        site=site_metadata,
        name="dummy fx",
        interval_length=pd.Timedelta("1h"),
        interval_value_type="instantaneous",
        issue_time_of_day='0500',
        lead_time_to_start=pd.Timedelta("1h"),
        run_length=pd.Timedelta("1h"),
        variable=variable,
        interval_label=interval_label,
    )

    assert fx.interval_label == "event"
    assert fx.variable == "event"


@pytest.mark.parametrize('model', [
    datamodel.Forecast,
    datamodel.EventForecast,
])
def test_event_forecast_from_dict(model, single_event_forecast):
    event_fx_dict = single_event_forecast.to_dict()
    assert isinstance(model.from_dict(event_fx_dict), datamodel.EventForecast)


def test_raw_report_event_forecast_loading(raw_report_dict_with_event):
    raw_report = datamodel.RawReport.from_dict(raw_report_dict_with_event)
    for fxobs in raw_report.processed_forecasts_observations:
        assert isinstance(fxobs.original.forecast, datamodel.EventForecast)


@pytest.mark.parametrize('model', [
    datamodel.Forecast,
    datamodel.ProbabilisticForecast,
])
def test_prob_forecast_from_dict(model, prob_forecasts):
    fx_dict = prob_forecasts.to_dict()
    assert isinstance(
        model.from_dict(fx_dict), datamodel.ProbabilisticForecast)


@pytest.mark.parametrize('model', [
    datamodel.Forecast,
    datamodel.ProbabilisticForecastConstantValue,
])
def test_prob_forecast_constant_value_from_dict(
        model, prob_forecast_constant_value):
    fx_dict = prob_forecast_constant_value.to_dict()
    assert isinstance(
        model.from_dict(fx_dict), datamodel.ProbabilisticForecastConstantValue)


def test_raw_report_prob_forecast_loading(raw_report_dict_with_prob):
    raw_report = datamodel.RawReport.from_dict(raw_report_dict_with_prob)
    assert isinstance(
        raw_report.processed_forecasts_observations[0].original.forecast,
        datamodel.ProbabilisticForecast)
    assert isinstance(
        raw_report.processed_forecasts_observations[1].original.forecast,
        datamodel.ProbabilisticForecastConstantValue)


@pytest.mark.parametrize('a,b', [
    ('constant', 'datetime'),
    ('constant', 'timeofday'),
    ('constant', 'errorband'),
    ('timeofday', 'constant'),
    ('timeofday', 'datetime'),
    ('timeofday', 'errorband'),
    ('datetime', 'constant'),
    ('datetime', 'timeofday'),
    ('datetime', 'errorband'),
    pytest.param('erorband', 'constant',
                 marks=pytest.mark.xfail(type=ValueError, strict=True)),
])
def test_invalid_costband(constant_cost, datetime_cost, timeofday_cost,
                          errorband_cost, a, b):
    dd = {'constant': constant_cost, 'timeofday': timeofday_cost,
          'datetime': datetime_cost, 'errorband': errorband_cost}
    with pytest.raises(TypeError):
        datamodel.CostBand(
            error_range=(0, 2),
            cost_function=a,
            cost_function_parameters=dd[b]
        )


@pytest.mark.parametrize('which', [
    'constant', 'timeofday', 'datetime', 'errorband',
    pytest.param('fullcost', marks=pytest.mark.xfail(
        type=ValueError, strict=True))
])
def test_cost_from_dict(cost_dicts, which):
    out = datamodel.Cost.from_dict({'name': 'test', 'type': which,
                                    'parameters': cost_dicts[which]})
    assert isinstance(out, datamodel.Cost)
    assert str(out.parameters).split('(')[0].lower() == which + 'cost'


@pytest.mark.parametrize('which', [
    'constant', 'timeofday', 'datetime', 'errorband',
    pytest.param('other', marks=pytest.mark.xfail(type=ValueError,
                                                  strict=True))
])
def test_cost_types(which, constant_cost, datetime_cost, timeofday_cost,
                    errorband_cost):
    dd = {'constant': constant_cost, 'timeofday': timeofday_cost,
          'datetime': datetime_cost, 'errorband': errorband_cost,
          'other': constant_cost}
    out = datamodel.Cost(
        name='test',
        type=which,
        parameters=dd[which]
    )
    assert isinstance(out, datamodel.Cost)
    assert str(out.parameters).split('(')[0].lower() == which + 'cost'


@pytest.mark.parametrize('which,cls,new', [
    ('constant', datamodel.ConstantCost, {'aggregation': 'min'}),
    ('timeofday', datamodel.TimeOfDayCost, {'aggregation': 'what'}),
    ('timeofday', datamodel.TimeOfDayCost, {'fill': 'no'}),
    ('timeofday', datamodel.TimeOfDayCost, {'times': tuple()}),
    ('timeofday', datamodel.TimeOfDayCost, {'cost': tuple()}),
    ('timeofday', datamodel.TimeOfDayCost, {'cost': tuple(), 'fill': 0}),
    ('datetime', datamodel.DatetimeCost, {'datetimes': tuple()}),
    ('datetime', datamodel.DatetimeCost, {'cost': tuple()}),
    ('datetime', datamodel.DatetimeCost, {'aggregation': tuple()}),
    ('datetime', datamodel.DatetimeCost, {'fill': 1})
])
def test_cost_validation(which, cls, new, cost_dicts):
    cd = cost_dicts[which]
    cd.update(new)
    with pytest.raises(ValueError):
        cls.from_dict(cd)


@pytest.mark.parametrize('param', [
    pytest.param(
        'wont be', marks=pytest.mark.xfail(strict=True, type=ValueError)),
    'example cost',
    None
])
def test_reportparameters_cost_consistency(report_params, param):
    nop = report_params.object_pairs[0].replace(cost=param)
    op = list(report_params.object_pairs) + [nop]
    datamodel.ReportParameters(
        name=report_params.name,
        start=report_params.start,
        end=report_params.end,
        object_pairs=op,
        metrics=('cost', 'mae'),
        categories=report_params.categories,
        filters=report_params.filters,
        costs=report_params.costs
    )


@pytest.fixture(params=['constant', 'timeofday', 'datetime', 'errorband'])
def cost_json(constant_cost, constant_cost_json, timeofday_cost,
              timeofday_cost_json, datetime_cost, datetime_cost_json,
              banded_cost_params, banded_cost_params_json, request):
    if request.param == 'constant':
        return constant_cost_json, datamodel.Cost(
            name='constantcost', type='constant', parameters=constant_cost)
    elif request.param == 'timeofday':
        return timeofday_cost_json, datamodel.Cost(
            name='timeofdaycost', type='timeofday', parameters=timeofday_cost)
    elif request.param == 'datetime':
        return datetime_cost_json, datamodel.Cost(
            name='datetimecost', type='datetime', parameters=datetime_cost)
    elif request.param == 'errorband':
        return banded_cost_params_json, banded_cost_params


def test_cost_from_json(cost_json):
    cjson, exp = cost_json
    out = datamodel.Cost.from_dict(json.loads(cjson))
    assert out == exp



from contextlib import contextmanager
from hashlib import sha256
import logging
import warnings


import numpy as np
import pandas as pd


from solarforecastarbiter import datamodel


def _observation_valid(index, obs_id, aggregate_observations):
    """
    Indicates where the observation data is valid. For now,
    effective_from and effective_until are inclusive, so data missing
    at those times is marked as missing in the aggregate.
    """
    nindex = pd.DatetimeIndex([], tz=index.tz)
    for aggobs in aggregate_observations:
        if aggobs['observation_id'] == obs_id:
            if aggobs['observation_deleted_at'] is None:
                locs = index.slice_locs(aggobs['effective_from'],
                                        aggobs['effective_until'])
                nindex = nindex.union(index[locs[0]:locs[1]])
            elif (
                    aggobs['effective_until'] is None or
                    aggobs['effective_until'] >= index[0]
            ):
                raise ValueError(
                    'Deleted Observation data cannot be retrieved'
                    ' to include in Aggregate')
            else:  # observation deleted and effective_until before index
                return pd.Series(False, index=index)
    return pd.Series(1, index=nindex).reindex(index).fillna(0).astype(bool)


def _make_aggregate_index(data, interval_length, interval_label,
                          timezone):
    """
    Compute the aggregate the index should have based on the min and
    max timestamps in the data, the interval length, label, and timezone.
    """
    # first, find limits for a new index
    start = pd.Timestamp('20380119T031407Z')
    end = pd.Timestamp('19700101T000001Z')
    for df in data.values():
        start = min(start, min(df.index))
        end = max(end, max(df.index))
    # adjust start, end to nearest interval
    # hard to understand what this interval should be for
    # odd (e.g. 52min) intervals, so required that interval
    # is a divisor of one day
    if 86400 % pd.Timedelta(interval_length).total_seconds() != 0:
        raise ValueError(
            'interval_length must be a divisor of one day')
    if interval_label == 'ending':
        start = start.ceil(interval_length)
        end = end.ceil(interval_length)
    elif interval_label == 'beginning':
        start = start.floor(interval_length)
        end = end.floor(interval_length)
    else:
        raise ValueError(
            'interval_label must be beginning or ending for aggregates')
    # raise the error if unlocalized
    start = start.tz_convert(timezone)
    end = end.tz_convert(timezone)
    return pd.date_range(
        start, end, freq=interval_length, tz=timezone)


def compute_aggregate(data, interval_length, interval_label,
                      timezone, agg_func, aggregate_observations):
    """
    Computes an aggregate quantity according to agg_func of the data.
    This function assumes the data has an interval_value_type of
    interval_mean or instantaneous and that the data interval_length
    is less than or equal to the aggregate interval_length.
    NaNs in the output are the result of missing data from an
    underyling observation of the aggregate.

    Parameters
    ----------
    data : dict of pandas.DataFrames
        With keys 'observation_id' corresponding to observation in
        aggregate_observations. DataFrames must have 'value' and 'quality_flag'
        columns.
    interval_length : str or pandas.Timedelta
        The time between timesteps in the aggregate result.
    interval_label : str
        Whether the timestamps in the aggregated output represent the beginning
        or ending of the interval
    timezone : str
        The IANA timezone for the output index
    agg_func : str
        The aggregation function (e.g 'sum', 'mean', 'min') to create the
        aggregate
    aggregate_observations : tuple of dicts
        Each dict should have 'observation_id' (string),
        'effective_from' (timestamp), 'effective_until' (timestamp or None),
        and 'observation_deleted_at' (timestamp or None) fields.

    Returns
    -------
    pandas.DataFrame
        - Index is a DatetimeIndex that adheres to interval_length and
          interval_label

        - Columns are 'value', for the aggregated value according to agg_func,
          and 'quality_flag', the bitwise or of all flags in the aggregate for
          the interval.

        - A 'value' of NaN means that data from one or more
          observations was missing in that interval.

    Raises
    ------
    KeyError
        If data is missing a key for an observation in aggregate_obsevations

        + Or, if any DataFrames in data do not have 'value' or 'quality_flag'
          columns

    ValueError
        If interval_length is not a divisor of one day

        + Or, if an observation has been deleted but the data is required for
          the aggregate
        + Or, if interval_label is not beginning or ending

    """
    new_index = _make_aggregate_index(
        data, interval_length, interval_label, timezone)
    unique_ids = {ao['observation_id'] for ao in aggregate_observations}
    valid_mask = {obs_id: _observation_valid(
        new_index, obs_id, aggregate_observations) for obs_id in unique_ids}

    missing_from_data_dict = {
        ao['observation_id'] for ao in aggregate_observations
        if ao['observation_deleted_at'] is None
        } - set(data.keys())

    if missing_from_data_dict:
        raise KeyError(
            'Cannot aggregate data with missing keys '
            f'{", ".join(missing_from_data_dict)}')

    value_is_missing = pd.Series(False, index=new_index)
    value = {}
    qf = {}
    closed = datamodel.CLOSED_MAPPING[interval_label]
    for obs_id, df in data.items():
        resampler = df.resample(interval_length, closed=closed, label=closed)
        new_val = resampler['value'].mean().reindex(new_index)
        # data is missing when the resampled value is NaN and the data
        # should be valid according to effective_from/until
        valid = valid_mask[obs_id]
        missing = new_val.isna() & valid
        if missing.any():
            warnings.warn('Values missing for one or more observations')
            value_is_missing[missing] = True
        value[obs_id] = new_val[valid]
        qf[obs_id] = resampler['quality_flag'].apply(np.bitwise_or.reduce)
    final_value = pd.DataFrame(value).reindex(new_index).aggregate(
        agg_func, axis=1)
    final_value[value_is_missing] = np.nan
    # have to fill in nans and convert to int to do bitwise_or
    # only works with pandas >= 0.25.0
    final_qf = pd.DataFrame(qf).reindex(new_index).fillna(0).astype(
        int).aggregate(np.bitwise_or.reduce, axis=1)
    out = pd.DataFrame({'value': final_value, 'quality_flag': final_qf})
    return out


def sha256_pandas_object_hash(obj):
    """
    Compute a hash for a pandas object. No sorting of the
    object is performed, so an object with the same data in
    in a different order returns a different hash.

    Parameters
    ----------
    obj: pandas.Series or pandas.DataFrame

    Returns
    -------
    str
       Hex digest of the SHA-256 hash of the individual object row hashes
    """
    return sha256(
        pd.util.hash_pandas_object(obj).values.tobytes()
    ).hexdigest()


class ListHandler(logging.Handler):
    """
    A logger handler that appends each log record to a list.
    """
    def __init__(self):
        super().__init__()
        self.records = []

    def emit(self, record):
        self.records.append(record)

    def export_records(self, level=logging.WARNING):
        """
        Convert each log record in the records list with level
        greater than or equal to `level` to a
        :py:class:`solarforecastarbiter.datamodel.ReportMessage`
        and return the tuple of messages.
        """
        out = []
        for rec in self.records:
            if rec.levelno >= level:
                out.append(
                    datamodel.ReportMessage(
                        message=rec.getMessage(),
                        step=rec.name,
                        level=rec.levelname,
                        function=rec.funcName
                    )
                )
        return tuple(out)


@contextmanager
def hijack_loggers(loggers, level=logging.INFO):
    """
    Context manager to temporarily set the handler
    of each logger in `loggers`.

    Parameters
    ----------
    loggers: list of str or logging.Logger
        Loggers to change
    level: logging LEVEL int
        Level to set the temporary handler to

    Returns
    -------
    ListHandler
        The handler that will be temporarily assigned
        to each logger.

    Notes
    -----
    This may not capture all records when used in a
    distributed or multiprocessing workflow
    """
    handler = ListHandler()
    handler.setLevel(level)

    old_handlers = {}
    for name in loggers:
        logger = logging.getLogger(name)
        old_handlers[name] = logger.handlers
        logger.handlers = [handler]
    yield handler
    for name in loggers:
        logger = logging.getLogger(name)
        logger.handlers = old_handlers[name]
    del handler


def _unique_key(try_key, keys, i=0):
    if try_key in keys:
        try_key += str(i)
        i += 1
        return _unique_key(try_key, keys, i)
    else:
        return try_key


def generate_continuous_chunks(data, freq):
    """
    Generator to split data into continuous chunks with spacing of freq.

    Parameters
    ----------
    data : pandas.Series or pandas.DataFrame
        Data to apply func to. Must have a DatetimeIndex.
    freq : pd.Timedelta
        Expected frequency to split data into continuous chunks

    Yields
    ------
    continuous_data : same as data
        Each continuous chunk that conforms to freq

    Raises
    ------
    TypeError
        If data is not a pandas Series or DataFrame, or
        does not have a DatetimeIndex
    ValueError
        If freq cannot be converted to a pandas.Timedelta

    Examples
    --------
    The following code would post two forecast series ignoring the missing
    period in the middle.

    .. testsetup::

        import pandas as pd
        from solarforecastarbiter.io import api

    >>> series = pd.Series(
    ...     [1.0, 2.0, 3.0, 7.0, 8.0],
    ...     index=[
    ...         pd.Timestamp('2020-07-01T01:00Z'),
    ...         pd.Timestamp('2020-07-01T02:00Z'),
    ...         pd.Timestamp('2020-07-01T03:00Z'),
    ...         pd.Timestamp('2020-07-01T07:00Z'),
    ...         pd.Timestamp('2020-07-01T08:00Z'),
    ...     ])
    >>> session = api.APISession('token')
    >>> for cser in generate_continuous_chunks(series, pd.Timedelta('1h')):
    ...     session.post_forecast_values('forecast_id', cser)
    """
    if (
            not isinstance(data, (pd.Series, pd.DataFrame)) or
            not isinstance(data.index, pd.DatetimeIndex)
    ):
        raise TypeError(
            'data must be a pandas Series or DataFrame with DatetimeIndex')
    if not isinstance(freq, pd.Timedelta):
        freq = pd.Timedelta(freq)

    # first value is NaT, rest are timedeltas
    index_ser = data.index.to_series()
    delta = index_ser.diff()
    if len(delta) < 2:
        yield data
        return

    flags = delta != freq

    group_key = '_cid'
    if isinstance(data, pd.DataFrame):
        keys = data.columns
        group_key = _unique_key(group_key, keys)
        data[group_key] = flags.cumsum()
    elif isinstance(data, pd.Series):
        keys = data.name
        data = pd.DataFrame(
            {keys: data, group_key: flags.cumsum()})

    for _, group in data.groupby(group_key):
        yield group[keys]



"""
Calculate AC power and modeling intermediates from system metadata,
times, and weather data.

Steps are:

1. Calculate solar position using solar_position
2. If not already known, calculate 3 irradiance components from measured
   GHI using irradiance_components or modeled clear sky using clearsky.
3. calculate_poa_effective
4. calculate_power

Steps 3 and 4 are bundled in :py:func:`irradiance_to_power`
"""

from functools import partial

import pvlib

from solarforecastarbiter import datamodel


def calculate_solar_position(latitude, longitude, elevation, times):
    """
    Calculates solar position using pvlib's implementation of NREL SPA.

    Parameters
    ----------
    latitude : float
    longitude : float
    elevation : float
    times : pd.DatetimeIndex

    Returns
    -------
    solar_position : pd.DataFrame
        The DataFrame will have the following columns: apparent_zenith
        (degrees), zenith (degrees), apparent_elevation (degrees),
        elevation (degrees), azimuth (degrees),
        equation_of_time (minutes).
    """
    solpos = pvlib.solarposition.get_solarposition(times, latitude,
                                                   longitude,
                                                   altitude=elevation,
                                                   method='nrel_numpy')
    return solpos


def complete_irradiance_components(ghi, zenith):
    """
    Uses the Erbs model to calculate DNI and DHI from GHI.

    Parameters
    ----------
    ghi : pd.Series
    zenith : pd.Series
        Solar zenith (not-refraction corrected)

    Returns
    -------
    dni : pd.Series, dhi : pd.Series
    """
    dni_dhi = pvlib.irradiance.erbs(ghi, zenith, ghi.index)
    return dni_dhi['dni'], dni_dhi['dhi']


def calculate_clearsky(latitude, longitude, elevation, apparent_zenith):
    """
    Calculates clear sky irradiance using the Ineichen model and the
    SoDa climatological turbidity data set.

    Parameters
    ----------
    latitude : float
    longitude : float
    elevation : float
    apparent_zenith : pd.Series
        Solar apparent zenith

    Returns
    -------
    cs : pd.DataFrame
        Columns are ghi, dni, dhi.
    """
    airmass = pvlib.atmosphere.get_relative_airmass(apparent_zenith)
    pressure = pvlib.atmosphere.alt2pres(elevation)
    am_abs = pvlib.atmosphere.get_absolute_airmass(airmass, pressure)
    tl = pvlib.clearsky.lookup_linke_turbidity(apparent_zenith.index,
                                               latitude,
                                               longitude)
    dni_extra = pvlib.irradiance.get_extra_radiation(apparent_zenith.index)
    cs = pvlib.clearsky.ineichen(apparent_zenith, am_abs, tl,
                                 dni_extra=dni_extra,
                                 altitude=elevation)
    return cs


def aoi_func_factory(modeling_parameters):
    """
    Create a function to calculate AOI, surface tilt, and surface
    azimuth from system modeling_parameters.

    Parameters
    ----------
    modeling_parameters : datamodel.FixedTiltModelingParameters or
                          datamodel.SingleAxisModelingParameters

    Returns
    -------
    function
        Function that accepts two arguments (apparent_zenith, azimuth)
        and returns three series (surface_tilt, surface_azimuth, aoi)

    Raises
    ------
    TypeError if modeling_parameters is invalid.
    """
    if isinstance(modeling_parameters,
                  datamodel.FixedTiltModelingParameters):
        return partial(
            aoi_fixed,
            modeling_parameters.surface_tilt,
            modeling_parameters.surface_azimuth
        )
    elif isinstance(modeling_parameters,
                    datamodel.SingleAxisModelingParameters):
        return partial(
            aoi_tracking,
            modeling_parameters.axis_tilt,
            modeling_parameters.axis_azimuth,
            modeling_parameters.max_rotation_angle,
            modeling_parameters.backtrack,
            modeling_parameters.ground_coverage_ratio
        )
    else:
        raise TypeError('Invalid modeling_parameters type %s' %
                        type(modeling_parameters))


def aoi_fixed(surface_tilt, surface_azimuth, apparent_zenith, azimuth):
    """
    Calculate AOI for fixed system, bundle return with tilt, azimuth for
    consistency with similar tracker function.

    Parameters
    ----------
    surface_tilt : float
    surface_azimuth : float
    apparent_zenith : pd.Series
        Solar apparent zenith
    azimuth : pd.Series
        Solar azimuth

    Returns
    -------
    surface_tilt : pd.Series, surface_azimuth : pd.Series, aoi : pd.Series
    """
    aoi = pvlib.irradiance.aoi(surface_tilt, surface_azimuth,
                               apparent_zenith, azimuth)
    return surface_tilt, surface_azimuth, aoi


def aoi_tracking(axis_tilt, axis_azimuth, max_rotation_angle, backtrack,
                 ground_coverage_ratio, apparent_zenith, azimuth):
    """
    Calculate AOI, surface tilt, and surface azimuth for tracking system.

    Parameters
    ----------
    axis_tilt : float
    axis_azimuth : float
    max_rotation_angle : float
    backtrack : bool
    ground_coverage_ratio : float
    apparent_zenith : pd.Series
        Solar apparent zenith
    azimuth : pd.Series
        Solar azimuth

    Returns
    -------
    surface_tilt : pd.Series, surface_azimuth : pd.Series, aoi : pd.Series
    """
    tracking = pvlib.tracking.singleaxis(
        apparent_zenith,
        azimuth,
        axis_tilt=axis_tilt,
        axis_azimuth=axis_azimuth,
        max_angle=max_rotation_angle,
        backtrack=backtrack,
        gcr=ground_coverage_ratio
    )
    surface_tilt = tracking['surface_tilt']
    surface_azimuth = tracking['surface_azimuth']
    aoi = tracking['aoi']
    return surface_tilt, surface_azimuth, aoi


def calculate_poa_effective_explicit(surface_tilt, surface_azimuth, aoi,
                                     apparent_zenith, azimuth, ghi, dni, dhi):
    """
    Calculate effective plane of array irradiance from system metadata,
    solar position, and irradiance components. Accounts for AOI losses.

    Parameters
    ----------
    surface_tilt : float or pd.Series
    surface_azimuth : float or pd.Series
    aoi : pd.Series
    apparent_zenith : pd.Series
        Solar apparent zenith
    azimuth : pd.Series
        Solar azimuth
    ghi : pd.Series
    dni : pd.Series
    dhi : pd.Series

    Returns
    -------
    poa_effective : pd.Series
    """
    dni_extra = pvlib.irradiance.get_extra_radiation(apparent_zenith.index)
    poa_sky_diffuse = pvlib.irradiance.get_sky_diffuse(
        surface_tilt, surface_azimuth,
        apparent_zenith, azimuth,
        dni, ghi, dhi,
        dni_extra=dni_extra,
        model='haydavies')
    poa_ground_diffuse = pvlib.irradiance.get_ground_diffuse(
        surface_tilt, ghi, albedo=0.25)
    aoi_modifier = pvlib.iam.physical(aoi)
    beam_effective = dni * aoi_modifier
    poa_effective = beam_effective + poa_sky_diffuse + poa_ground_diffuse
    # aoi, tilt, azi is not defined for tracking systems
    # when sun is below horizon. replace nan with 0
    poa_effective = poa_effective.where(aoi.notna(), other=0.)
    return poa_effective


def calculate_poa_effective(aoi_func, apparent_zenith, azimuth, ghi, dni, dhi):
    """
    Calculate effective plane of array irradiance from system metadata,
    solar position, and irradiance components. Accounts for AOI losses.

    Parameters
    ----------
    aoi_func : function
        Function with arguments (apparent_zenith, azimuth) and returns
        surface_tilt, surface_azimuth, aoi
    apparent_zenith : pd.Series
        Solar apparent zenith
    azimuth : pd.Series
        Solar azimuth
    ghi : pd.Series
    dni : pd.Series
    dhi : pd.Series

    Returns
    -------
    poa_effective : pd.Series
    """
    surface_tilt, surface_azimuth, aoi = aoi_func(apparent_zenith, azimuth)
    poa_effective = calculate_poa_effective_explicit(
        surface_tilt, surface_azimuth, aoi, apparent_zenith, azimuth,
        ghi, dni, dhi)
    return poa_effective


def calculate_power(dc_capacity, temperature_coefficient, dc_loss_factor,
                    ac_capacity, ac_loss_factor, poa_effective, temp_air=20,
                    wind_speed=1):
    """
    Calcuate AC power from system metadata, plane of array irradiance,
    and weather data using the PVWatts model.

    Parameters
    ----------
    dc_capacity : float
    temperature_coefficient : float
        Specified in units of %/C to be converted to 1/C
    dc_loss_factor : float
    ac_capacity : float
    ac_loss_factor : float
    poa_effective : pd.Series
    temp_air : pd.Series, default 20
    wind_speed : pd.Series, default 1

    Returns
    -------
    ac_power : pd.Series
    """
    pvtemps = pvlib.temperature.pvsyst_cell(poa_effective, temp_air,
                                            wind_speed=wind_speed)
    dc = pvlib.pvsystem.pvwatts_dc(poa_effective, pvtemps, dc_capacity,
                                   temperature_coefficient / 100)
    dc *= (1 - dc_loss_factor / 100)
    # set eta values to turn off clipping in pvwatts inverter model
    ac = pvlib.inverter.pvwatts(dc, dc_capacity, eta_inv_nom=1, eta_inv_ref=1)
    ac = ac.clip(upper=ac_capacity)
    ac *= (1 - ac_loss_factor / 100)
    return ac


def irradiance_to_power(modeling_parameters, apparent_zenith, azimuth, ghi,
                        dni, dhi, temp_air=20, wind_speed=1):
    """
    Calcuate AC power from system metadata, solar position, and
    ghi, dni, dhi.

    Parameters
    ----------
    modeling_parameters : datamodel.FixedTiltModelingParameters or
                          datamodel.SingleAxisModelingParameters
    apparent_zenith : pd.Series
        Solar apparent zenith
    azimuth : pd.Series
        Solar azimuth
    ghi : pd.Series
    dni : pd.Series
    dhi : pd.Series
    temp_air : pd.Series, default 20
    wind_speed : pd.Series, default 1

    Returns
    -------
    ac_power : pd.Series
    """
    aoi_func = aoi_func_factory(modeling_parameters)
    poa_effective = calculate_poa_effective(
        aoi_func, apparent_zenith, azimuth, ghi, dni, dhi)
    ac = calculate_power(
        modeling_parameters.dc_capacity,
        modeling_parameters.temperature_coefficient,
        modeling_parameters.dc_loss_factor,
        modeling_parameters.ac_capacity,
        modeling_parameters.ac_loss_factor,
        poa_effective,
        temp_air=temp_air,
        wind_speed=wind_speed)
    return ac



import logging
from urllib.error import URLError


import pandas as pd
from pvlib import iotools


from solarforecastarbiter.io.reference_observations import (
    common, default_forecasts)


solrad_variables = ['ghi', 'dni', 'dhi']

SOLRAD_FTP_DIR = "ftp://aftp.cmdl.noaa.gov/data/radiation/solrad/"
REALTIME_URL = SOLRAD_FTP_DIR + "/realtime/{abbr}/{abrv}{year_2d}{jday}.dat"
ARCHIVE_URL = SOLRAD_FTP_DIR + "/{abrv}/{year}/{abrv}{year_2d}{jday}.dat"

logger = logging.getLogger('reference_data')


def fetch(api, site, start, end, realtime=False):
    """Retrieve observation data for a solrad site between start and end.

    Parameters
    ----------
    api : io.APISession
        An APISession with a valid JWT for accessing the Reference Data
        user.
    site : datamodel.Site
        Site object with the appropriate metadata.
    start : datetime
        The beginning of the period to request data for.
    end : datetime
        The end of the period to request data for.
    realtime : bool
        Whether or not to look for realtime data. Note that this data is
        raw, unverified data from the instruments.

    Returns
    -------
    data : pandas.DataFrame
        All of the requested data concatenated into a single DataFrame.
    """
    if realtime:
        url_format = REALTIME_URL
    else:
        url_format = ARCHIVE_URL
    # load extra parameters for api arguments.
    extra_params = common.decode_extra_parameters(site)
    abbreviation = extra_params['network_api_abbreviation']
    single_day_dfs = []
    for day in pd.date_range(start, end):
        filename = url_format.format(abrv=abbreviation,
                                     year=day.year,
                                     year_2d=day.strftime('%y'),
                                     jday=day.strftime('%j'))
        logger.info(f'Requesting data for SOLRAD site {site.name}'
                    f' on {day.strftime("%Y%m%d")}.')
        try:
            # Only get dataframe from the returned tuple
            solrad_day = iotools.read_solrad(filename)
        except URLError:
            logger.warning(f'Could not retrieve SOLRAD data for site '
                           f'{site.name} on {day.strftime("%Y%m%d")}.')
            logger.debug(f'Failed SOLRAD URL: {filename}.')
            continue
        else:
            single_day_dfs.append(solrad_day)
    try:
        all_period_data = pd.concat(single_day_dfs)
    except ValueError:
        logger.warning(f'No data available for site {site.name} '
                       f'from {start} to {end}.')
        return pd.DataFrame()
    return all_period_data


def initialize_site_observations(api, site):
    """Creates an observaiton at the site for each variable in solrad_variables.

    Parameters
    ----------
    site : datamodel.Site
        The site object for which to create Observations.
    """
    for variable in solrad_variables:
        common.create_observation(api, site, variable)


def initialize_site_forecasts(api, site):
    """
    Create a forecasts for each variable measured at the site

    Parameters
    ----------
    api : solarforecastarbiter.io.api.APISession
        An active Reference user session.
    site : datamodel.Site
        The site object for which to create Forecasts.
    """
    common.create_forecasts(api, site, solrad_variables,
                            default_forecasts.TEMPLATE_FORECASTS)


def update_observation_data(api, sites, observations, start, end):
    """Post new observation data to all reference observations
    at each SOLRAD site between start and end.

    api : solarforecastarbiter.io.api.APISession
        An active Reference user session.
    sites : list of solarforecastarbiter.datamodel.Site
        List of all reference sites as Objects
    observations : list of solarforecastarbiter.datamodel.Observation
        List of all reference observations.
    start : datetime
        The beginning of the period to request data for.
    end : datetime
        The end of the period to request data for.
    """
    solrad_sites = common.filter_by_networks(sites, 'NOAA SOLRAD')
    for site in solrad_sites:
        common.update_site_observations(
            api, fetch, site, observations, start, end)



"""
Metric calculation functions.
"""
import calendar
from functools import partial
import logging

import numpy as np
import pandas as pd


from solarforecastarbiter import datamodel
from solarforecastarbiter.metrics import deterministic, probabilistic, event


logger = logging.getLogger(__name__)


def calculate_metrics(processed_pairs, categories, metrics):
    """
    Loop through the forecast-observation pairs and calculate metrics.

    Normalization is determined by the attributes of the input objects.

    If ``processed_fx_obs.uncertainty`` is not ``None``, a deadband
    equal to the uncertainty will be used by the metrics that support it.

    Parameters
    ----------
    processed_pairs :
        List of datamodel.ProcessedForecastObservation
    categories : list of str
        List of categories to compute metrics over.
    metrics : list of str
        List of metrics to be computed.

    Returns
    -------
    list
        List of datamodel.MetricResult for each
        datamodel.ProcessedForecastObservation
    """
    calc_metrics = []

    for proc_fxobs in processed_pairs:

        # determine type of metrics to calculate
        if isinstance(proc_fxobs.original.forecast,
                      (datamodel.ProbabilisticForecast,
                       datamodel.ProbabilisticForecastConstantValue)):
            try:
                calc_metrics.append(calculate_probabilistic_metrics(
                    proc_fxobs,
                    categories,
                    metrics))
            except RuntimeError as e:
                logger.error('Failed to calculate probabilistic metrics'
                             ' for %s: %s', proc_fxobs.name, e)
        elif isinstance(proc_fxobs.original.forecast, datamodel.EventForecast):
            try:
                calc_metrics.append(calculate_event_metrics(
                    proc_fxobs, categories, metrics
                ))
            except RuntimeError as e:
                logger.error('Failed to calculate event metrics for %s: %s',
                             proc_fxobs.name, e)
        else:
            try:
                calc_metrics.append(calculate_deterministic_metrics(
                    proc_fxobs,
                    categories,
                    metrics))
            except RuntimeError as e:
                logger.error('Failed to calculate deterministic metrics'
                             ' for %s: %s',
                             proc_fxobs.name, e)

    return calc_metrics


def _apply_deterministic_metric_func(metric, fx, obs, **kwargs):
    """Helper function to deal with variable number of arguments possible for
    deterministic metric functions."""
    metric_func = deterministic._MAP[metric][0]

    # the keyword arguments that will be passed to the functions
    _kw = {}

    # deadband could be supplied as None, so this is a little cleaner
    # than a try/except pattern
    deadband = kwargs.get('deadband', None)
    if metric in deterministic._DEADBAND_ALLOWED and deadband:
        # metrics assumes fractional deadband, datamodel assumes %, so / 100
        deadband_frac = deadband / 100
        _kw['error_fnc'] = partial(
            deterministic.error_deadband, deadband=deadband_frac)

    if metric == 'cost':
        _kw['cost_params'] = kwargs['cost_params']

    # ref is an arg, but seems cleaner to handle as a kwarg here
    if metric in deterministic._REQ_REF_FX:
        _kw['ref'] = kwargs['ref_fx']

    # same arg/kwarg comment as for ref
    if metric in deterministic._REQ_NORM:
        _kw['norm'] = kwargs['normalization']

    return metric_func(obs, fx, **_kw)


def _apply_probabilistic_metric_func(metric, fx, fx_prob, obs, **kwargs):
    """Helper function to deal with variable number of arguments possible for
    probabilistic metric functions."""
    metric_func = probabilistic._MAP[metric][0]
    if metric in probabilistic._REQ_REF_FX:
        return metric_func(obs, fx, fx_prob,
                           kwargs['ref_fx'], kwargs['ref_fx_prob'])
    else:
        return metric_func(obs, fx, fx_prob)


def _apply_event_metric_func(metric, fx, obs, **kwargs):
    """Helper function to deal with variable number of arguments possible for
    event metric functions."""
    metric_func = event._MAP[metric][0]
    return metric_func(obs, fx)


def calculate_deterministic_metrics(processed_fx_obs, categories, metrics):
    """
    Calculate deterministic metrics for the processed data using the provided
    categories and metric types.

    Normalization is determined by the attributes of the input objects.

    If ``processed_fx_obs.uncertainty`` is not ``None``, a deadband
    equal to the uncertainty will be used by the metrics that support it.

    Parameters
    ----------
    processed_fx_obs : datamodel.ProcessedForecastObservation
    categories : list of str
        List of categories to compute metrics over.
    metrics : list of str
        List of metrics to be computed.

    Returns
    -------
    solarforecastarbiter.datamodel.MetricResult
        Contains all the computed metrics by categories.

    Raises
    ------
    RuntimeError
        If there is no forecast, observation timeseries data or no metrics
        are specified.
    """
    out = {
        'name': processed_fx_obs.name,
        'forecast_id': processed_fx_obs.original.forecast.forecast_id,
    }

    try:
        out['observation_id'] = processed_fx_obs.original.observation.observation_id  # NOQA: E501
    except AttributeError:
        out['aggregate_id'] = processed_fx_obs.original.aggregate.aggregate_id

    fx = processed_fx_obs.forecast_values
    obs = processed_fx_obs.observation_values

    # Check reference forecast is from processed pair, if needed
    ref_fx = processed_fx_obs.reference_forecast_values
    if ref_fx is None:
        ref_fx = np.nan  # avoids issues with None and deadband masking

    # No data or metrics
    if fx.empty:
        raise RuntimeError("No forecast timeseries data.")
    elif obs.empty:
        raise RuntimeError("No observation timeseries data.")
    elif len(metrics) == 0:
        raise RuntimeError("No metrics specified.")

    # Dataframe for grouping
    df = pd.DataFrame({'forecast': fx,
                       'observation': obs,
                       'reference': ref_fx})

    # get normalization factor
    normalization = processed_fx_obs.normalization_factor

    # get uncertainty.
    deadband = processed_fx_obs.uncertainty

    cost_params = processed_fx_obs.cost

    # Force `groupby` to be consistent with `interval_label`, i.e., if
    # `interval_label == ending`, then the last interval should be in the bin
    if processed_fx_obs.interval_label == "ending":
        df.index -= pd.Timedelta("1ns")

    metric_vals = []
    # Calculate metrics
    for category in set(categories):
        # total (special category)
        if category == 'total':
            index_category = lambda x: 0  # NOQA: E731
        else:
            index_category = getattr(df.index, category)

        # Calculate each metric
        for metric_ in metrics:

            # Group by category
            for cat, group in df.groupby(index_category):

                # Calculate
                res = _apply_deterministic_metric_func(
                    metric_, group.forecast, group.observation,
                    ref_fx=group.reference, normalization=normalization,
                    deadband=deadband, cost_params=cost_params)

                # Change category label of the group from numbers
                # to e.g. January or Monday
                if category == 'month':
                    cat = calendar.month_abbr[cat]
                elif category == 'weekday':
                    cat = calendar.day_abbr[cat]

                metric_vals.append(datamodel.MetricValue(
                    category, metric_, str(cat), res))

    out['values'] = _sort_metrics_vals(
        metric_vals, datamodel.ALLOWED_DETERMINISTIC_METRICS)
    calc_metrics = datamodel.MetricResult.from_dict(out)
    return calc_metrics


def _sort_metrics_vals(metrics_vals, mapping):
    """
    Parameters
    ----------
    metrics_vals : list
        Elements are datamodel.MetricValue
    mapping : dict
        Metrics mapping defined in datamodel.

    Returns
    -------
    tuple
        Sorted elements. Sorting order is:
            * Category
            * Metric
            * Index
            * Value
    """
    metric_ordering = list(mapping.keys())
    category_ordering = list(datamodel.ALLOWED_CATEGORIES.keys())

    def sorter(metricval):
        if metricval.category == 'month':
            index_order = calendar.month_abbr[0:13].index(metricval.index)
        elif metricval.category == 'weekday':
            index_order = calendar.day_abbr[0:7].index(metricval.index)
        else:
            index_order = metricval.index

        return (
            category_ordering.index(metricval.category),
            metric_ordering.index(metricval.metric),
            index_order,
            metricval.value
            )

    metrics_sorted = tuple(sorted(metrics_vals, key=sorter))
    return metrics_sorted


def calculate_probabilistic_metrics(processed_fx_obs, categories, metrics):
    """
    Calculate probabilistic metrics for the processed data using the provided
    categories and metric types. Will calculate distribution metrics (e.g.,
    Continuous Ranked Probability Score (CRPS)) over all constant values if
    forecast is a datamodel.ProbabilisticForecast is provided, or single value
    metrics (e.g., Briar Score (BS)) if forecast is a
    datamodel.ProbabilisticForecastConstantValue.

    Parameters
    ----------
    processed_fx_obs : datamodel.ProcessedForecastObservation
        Forecasts must be datamodel.ProbabilisticForecast (CDFs) or
        datamodel.ProbabilisticForecastConstantValue (single values)
    categories : list of str
        List of categories to compute metrics over.
    metrics : list of str
        List of metrics to be computed.

    Returns
    -------
    datamodel.MetricsResult

    Raises
    ------
    RuntimeError
        If there is no forecast, observation timeseries data or no metrics
        are specified.
    ValueError
        If original and reference forecast ``axis`` values do not match.
    """
    out = {'name': processed_fx_obs.name}
    try:
        obs_dict = {'observation_id':
                    processed_fx_obs.original.observation.observation_id}
    except AttributeError:
        obs_dict = {'aggregate_id':
                    processed_fx_obs.original.aggregate.aggregate_id}
    out.update(obs_dict)

    # Determine forecast physical and probability values by axis
    fx_fx_prob = _transform_prob_forecast_value_and_prob(processed_fx_obs)
    obs = processed_fx_obs.observation_values

    if processed_fx_obs.reference_forecast_values is None:
        ref_fx_fx_prob = [(np.nan, np.nan)]*len(fx_fx_prob)
    else:
        if (processed_fx_obs.original.forecast.axis !=
                processed_fx_obs.original.reference_forecast.axis):
            # could put this check in datamodel (and others from preprocessing)
            raise ValueError("Mismatched `axis` between "
                             "forecast and reference forecast.")
        ref_fx_fx_prob = _transform_prob_forecast_value_and_prob(
            processed_fx_obs, attr='reference_forecast_values')
        out.update({'reference_forecast_id': processed_fx_obs.original.reference_forecast.forecast_id})  # NOQA: E501

    # No data or metrics
    if (not fx_fx_prob or
            any([fx[0].empty or fx[1].empty for fx in fx_fx_prob])):
        raise RuntimeError("Missing probabilistic forecast timeseries data.")
    elif obs.empty:
        raise RuntimeError("No observation timeseries data.")
    elif len(metrics) == 0:
        raise RuntimeError("No metrics specified.")

    fx = processed_fx_obs.original.forecast
    out['forecast_id'] = fx.forecast_id

    # Determine proper metrics by instance type
    if isinstance(fx, datamodel.ProbabilisticForecastConstantValue):
        _metrics = list(set(metrics) - set(probabilistic._REQ_DIST))
    elif isinstance(fx, datamodel.ProbabilisticForecast):
        _metrics = list(set(metrics) & set(probabilistic._REQ_DIST))

    # Compute metrics and create MetricResult
    results = _calculate_probabilistic_metrics_from_df(
        _create_prob_dataframe(obs, fx_fx_prob, ref_fx_fx_prob),
        categories,
        _metrics,
        processed_fx_obs.original.forecast.interval_label)
    out['values'] = _sort_metrics_vals(
        results, datamodel.ALLOWED_PROBABILISTIC_METRICS)
    result = datamodel.MetricResult.from_dict(out)
    return result


def _calculate_probabilistic_metrics_from_df(data_df, categories, metrics,
                                             interval_label):
    """
    Calculate probabilistic metrics for the processed data using the provided
    categories and metric types.

    Parameters
    ----------
    data_df : pandas.DataFrame
        DataFrame that contains all timeseries values on the same index.
    categories : list of str
        List of categories to compute metrics over.
    metrics : list of str
        List of metrics to be computed.
    interval_label : str

    Returns
    -------
    list of tuples of datamodel.MetricValue
        Contains all the computed metrics by categories. Each tuple is
        associated with a datamodel.ProbabilisticForecastConstantValue.
    """
    metric_values = []

    # Force `groupby` to be consistent with `interval_label`, i.e., if
    # `interval_label == ending`, then the last interval should be in the
    # bin
    if interval_label == "ending":
        data_df.index -= pd.Timedelta("1ns")

    # Calculate metrics
    for category in set(categories):
        # total (special category)
        if category == 'total':
            index_category = lambda x: 0  # NOQA: E731
        else:
            index_category = getattr(data_df.index, category)

        # Calculate each metric
        for metric_ in set(metrics):
            # Group by category
            for cat, group in data_df.groupby(index_category):

                try:
                    ref_fx_vals = group.xs('reference_forecast', level=1, axis=1).to_numpy()  # NOQA: E501
                    ref_fx_prob_vals = group.xs('reference_probability', level=1, axis=1).to_numpy()  # NOQA E501
                    if ref_fx_vals.size == ref_fx_vals.shape[0]:
                        ref_fx_vals = ref_fx_vals.T[0]
                        ref_fx_prob_vals = ref_fx_prob_vals.T[0]
                except KeyError:
                    ref_fx_vals = np.nan
                    ref_fx_prob_vals = np.nan

                fx_vals = group.xs('forecast', level=1, axis=1).to_numpy()
                fx_prob_vals = group.xs('probability', level=1, axis=1).to_numpy()  # NOQA: E501
                if fx_vals.size == fx_vals.shape[0]:
                    fx_vals = fx_vals.T[0]
                    fx_prob_vals = fx_prob_vals.T[0]
                obs_vals = group[(None, 'observation')].to_numpy()

                # Calculate
                res = _apply_probabilistic_metric_func(
                    metric_, fx_vals, fx_prob_vals, obs_vals,
                    ref_fx=ref_fx_vals, ref_fx_prob=ref_fx_prob_vals)

                # Change category label of the group from numbers
                # to e.g. January or Monday
                if category == 'month':
                    cat = calendar.month_abbr[cat]
                elif category == 'weekday':
                    cat = calendar.day_abbr[cat]

                metric_values.append(datamodel.MetricValue(
                    category, metric_, str(cat), res))

    return metric_values


def _create_prob_dataframe(obs, fx_fx_prob, ref_fx_fx_prob):
    """
    Creates a DataFrame for grouping for the probabilistic forecasts.

    Parameters
    ----------
    obs : pandas.Series
    fx_fx_prob : list of tuple of pandas.Series
    ref_fx_fx_prob : list of tuple of pandas.Series

    Returns
    -------
    pandas.DataFrame
        DataFrame combining all data with column names being tuples of
        constant_value and one of 'observation', 'forecast', 'probability',
        'reference_forecast' or 'reference_probability'. For 'observation' the
        constant_value is None.
    """
    data_dict = {(None, 'observation'): obs}

    for orig, ref in zip(fx_fx_prob, ref_fx_fx_prob):
        fx, fx_prob = orig
        ref_fx, ref_fx_prob = ref
        data_dict[(fx.name, 'forecast')] = fx
        data_dict[(fx_prob.name, 'probability')] = fx_prob
        try:
            data_dict[(ref_fx.name, 'reference_forecast')] = ref_fx
            data_dict[(ref_fx_prob.name, 'reference_probability')] = ref_fx_prob  # NOQA: E501
        except AttributeError:
            pass

    return pd.DataFrame(data_dict)


def _transform_prob_forecast_value_and_prob(proc_fx_obs,
                                            attr='forecast_values'):
    """
    Helper function that returns ordered list of series of physical values and
    probabilities for a datamodel.ProcessedForecastObservation

    Parameters
    ----------
    proc_fx_obs : datamodel.ProcessedForecastObservation
    attr : str
        The attribute of proc_fx_obs that contains the forecast value of
        interest. Typically forecast_values or
        reference_forecast_values.

    Returns
    -------
    fx_fx_prob : list of tuples of (n,) array_like
        Forecast physical values and forecast probabilities.
    """
    fx_fx_prob = []
    data = getattr(proc_fx_obs, attr)

    # Need to convert to DataFrame of ProbabilisticForecastConstantValue
    # for consistency with ProbabilisticForecas that provides a DataFrame
    if isinstance(data, pd.Series):
        # set the name to a float because it will default to 'value'
        data = data.to_frame(
            name=proc_fx_obs.original.forecast.constant_value)

    for col in data.columns:
        if proc_fx_obs.original.forecast.axis == 'x':
            fx_prob = data[col]
            fx = pd.Series([float(col)]*fx_prob.size, index=fx_prob.index,
                           name=col)
        else:  # 'y'
            fx = data[col]
            fx_prob = pd.Series([float(col)]*fx.size, index=fx.index,
                                name=col)
        fx_fx_prob.append((fx, fx_prob))
    return fx_fx_prob


def calculate_event_metrics(proc_fx_obs, categories, metrics):
    """
    Calculate event metrics for the processed data using the provided
    categories and metric types.

    Parameters
    ----------
    proc_fx_obs : datamodel.ProcessedForecastObservation
    categories : list of str
        List of categories to compute metrics over.
    metrics : list of str
        List of metrics to be computed.

    Returns
    -------
    solarforecastarbiter.datamodel.MetricResult
        Contains all the computed metrics by categories.

    Raises
    ------
    RuntimeError
        If there is no forecast, observation timeseries data or no metrics
        are specified.
    """

    out = {
        'name': proc_fx_obs.name,
        'forecast_id': proc_fx_obs.original.forecast.forecast_id,
        'observation_id': proc_fx_obs.original.observation.observation_id
    }

    fx = proc_fx_obs.forecast_values
    obs = proc_fx_obs.observation_values

    # No data or metrics
    if fx.empty:
        raise RuntimeError("No Forecast timeseries data.")
    elif obs.empty:
        raise RuntimeError("No Observation timeseries data.")
    elif len(metrics) == 0:
        raise RuntimeError("No metrics specified.")

    # Dataframe for grouping
    df = pd.concat({'forecast': fx, 'observation': obs}, axis=1)

    metric_vals = []
    # Calculate metrics
    for category in set(categories):
        # total (special category)
        if category == 'total':
            index_category = lambda x: 0  # NOQA
        else:
            index_category = getattr(df.index, category)

        # Calculate each metric
        for metric_ in set(metrics):
            # Group by category
            for cat, group in df.groupby(index_category):

                # Calculate
                res = _apply_event_metric_func(
                    metric_, group.forecast, group.observation
                )

                # Change category label of the group from numbers
                # to e.g. January or Monday
                if category == 'month':
                    cat = calendar.month_abbr[cat]
                elif category == 'weekday':
                    cat = calendar.day_abbr[cat]

                metric_vals.append(datamodel.MetricValue(
                    category, metric_, str(cat), res))

    out['values'] = _sort_metrics_vals(metric_vals,
                                       datamodel.ALLOWED_EVENT_METRICS)
    calc_metrics = datamodel.MetricResult.from_dict(out)
    return calc_metrics



import re

import pandas as pd
import pandas.testing as pdt

import pytest

from solarforecastarbiter.datamodel import (
    SolarPowerPlant, FixedTiltModelingParameters)
from solarforecastarbiter.io import api
from solarforecastarbiter.io.reference_observations import pvdaq


@pytest.fixture
def session(requests_mock):
    return api.APISession('')


@pytest.fixture
def site():
    return SolarPowerPlant(
        name='NREL x-Si #1',
        latitude=39.7406,
        longitude=-105.1774,
        elevation=1785.050170898438,
        timezone='Etc/GMT+7',
        site_id='',
        provider='',
        extra_parameters='{"network_api_id": 4, "attribution": "https://developer.nrel.gov/docs/solar/pvdaq-v3/", "network": "NREL PVDAQ", "network_api_abbreviation": "pvdaq", "observation_interval_length": 1, "inverter_mfg": "SMA", "inverter_model": "1800", "module_mfg": "Sanyo", "module_model": "HIP 200-BA3", "module_tech": "1"}',  # noqa: E501
        modeling_parameters=FixedTiltModelingParameters(
            ac_capacity=0.001,
            dc_capacity=0.001,
            temperature_coefficient=0.3,
            dc_loss_factor=0,
            ac_loss_factor=0,
            surface_tilt=40.0,
            surface_azimuth=180.0,
            tracking_type='fixed'))


@pytest.fixture
def site_no_extra():
    return SolarPowerPlant(
        name='NREL x-Si #1',
        latitude=39.7406,
        longitude=-105.1774,
        elevation=1785.050170898438,
        timezone='Etc/GMT+7',
        site_id='',
        provider='',
        extra_parameters='',
        modeling_parameters=FixedTiltModelingParameters(
            ac_capacity=0.001,
            dc_capacity=0.001,
            temperature_coefficient=0.3,
            dc_loss_factor=0,
            ac_loss_factor=0,
            surface_tilt=40.0,
            surface_azimuth=180.0,
            tracking_type='fixed'))


@pytest.fixture
def log(mocker):
    log = mocker.patch('solarforecastarbiter.io.reference_observations.'
                       'pvdaq.logger')
    return log


@pytest.fixture()
def mock_list_sites(mocker, many_sites):
    mocker.patch('solarforecastarbiter.io.api.APISession.list_sites',
                 return_value=many_sites)


def test_initialize_site_observations(
        requests_mock, mocker, session, site, single_observation,
        single_observation_text, mock_list_sites):
    matcher = re.compile(f'{session.base_url}/observations/.*')
    requests_mock.register_uri('POST', matcher,
                               text=single_observation.observation_id)
    requests_mock.register_uri('GET', matcher, content=single_observation_text)
    status = mocker.patch(
        'solarforecastarbiter.io.api.APISession.create_observation')
    pvdaq.initialize_site_observations(session, site)
    assert status.called


def test_initialize_site_observations_fail(session, site_no_extra, log):
    pvdaq.initialize_site_observations(session, site_no_extra)
    assert log.warning.call_count == 1


def test_initialize_site_forecasts(
        requests_mock, mocker, session, site, single_forecast,
        single_forecast_text, mock_list_sites):
    matcher = re.compile(f'{session.base_url}/forecasts/.*')
    requests_mock.register_uri('POST', matcher,
                               text=single_forecast.forecast_id)
    requests_mock.register_uri('GET', matcher, content=single_forecast_text)
    status = mocker.patch(
        'solarforecastarbiter.io.reference_observations.common.'
        'create_forecasts')
    pvdaq.initialize_site_forecasts(session, site)
    assert status.called


def test_initialize_site_forecasts_fail(session, site_no_extra, log):
    pvdaq.initialize_site_forecasts(session, site_no_extra)
    assert log.warning.call_count == 1


def test_fetch(mocker, session, site):
    status = mocker.patch(
        'solarforecastarbiter.io.fetch.pvdaq.get_pvdaq_data'
    )
    start = pd.Timestamp('2020-01-01T0000Z')
    end = pd.Timestamp('2020-01-02T0000Z')
    api_key = 'nopethepope'
    pvdaq.fetch(session, site, start, end, nrel_pvdaq_api_key=api_key)
    assert status.called


def test_fetch_fail(session, site_no_extra):
    start = pd.Timestamp('2020-01-01T0000Z')
    end = pd.Timestamp('2020-01-02T0000Z')
    api_key = 'nopethepope'
    out = pvdaq.fetch(session, site_no_extra, start, end,
                      nrel_pvdaq_api_key=api_key)
    assert out.empty


def test_fetch_fail_except(session, site, mocker, log):
    start = pd.Timestamp('2020-01-01T0000Z')
    end = pd.Timestamp('2020-01-02T0000Z')
    api_key = 'nopethepope'
    status = mocker.patch(
        'solarforecastarbiter.io.fetch.pvdaq.get_pvdaq_data'
    )
    status.side_effect = Exception
    out = pvdaq.fetch(session, site, start, end,
                      nrel_pvdaq_api_key=api_key)
    assert log.warning.call_count == 1
    assert out.empty


def test_fetch_fail_nonexistenttime(session, site, mocker, log):
    site = site.replace(timezone='America/Denver')
    start = pd.Timestamp('2020-01-01T0000Z')
    end = pd.Timestamp('2020-01-02T0000Z')
    api_key = 'nopethepope'
    index = pd.DatetimeIndex(['2020-03-08 02:00:00'])
    df = pd.DataFrame({'ac_power': 0}, index=index)
    patch = mocker.patch('solarforecastarbiter.io.fetch.pvdaq.get_pvdaq_data')
    patch.return_value = df
    out = pvdaq.fetch(session, site, start, end,
                      nrel_pvdaq_api_key=api_key)
    assert log.warning.call_count == 1
    assert out.empty


@pytest.fixture
def mock_pvdaq_creds(mocker):
    mocker.patch.dict('os.environ', {'NREL_PVDAQ_API_KEY': 'fake_key'})


def test_update_observation_data(mocker, session, site, mock_pvdaq_creds):
    obs_update = mocker.patch(
        'solarforecastarbiter.io.reference_observations.common.'
        'update_site_observations')
    start = pd.Timestamp('20200101T0000Z')
    end = pd.Timestamp('20200102T0000Z')
    pvdaq.update_observation_data(session, [site], [], start, end)
    obs_update.assert_called()
    assert obs_update.call_count == 1


def test_update_observation_data_no_creds(session, site):
    start = pd.Timestamp('20200101T0000Z')
    end = pd.Timestamp('20200102T0000Z')
    with pytest.raises(KeyError) as e:
        pvdaq.update_observation_data(session, [site], [], start, end)
    assert 'environment variable' in str(e.value)


def test_watts_to_mw():
    columns = [
        'ac_power', 'inv1_ac_power', 'AC_power', 'AC_Power', 'DC_power',
        'wind_speed', 'bla h_power_blob']
    obs_df = pd.DataFrame(
        1.e6, columns=columns, index=[pd.Timestamp('20200514')])
    expected = pd.DataFrame(
        [[1., 1., 1., 1., 1., 1.e6, 1.]], columns=columns,
        index=[pd.Timestamp('20200514')])
    out = pvdaq._watts_to_mw(obs_df)
    pdt.assert_frame_equal(out, expected)



import logging
import json
from urllib import error
from pkg_resources import resource_filename, Requirement


import pandas as pd
from pvlib import iotools
from requests.exceptions import HTTPError


from solarforecastarbiter.datamodel import Observation, SolarPowerPlant
from solarforecastarbiter.io.reference_observations import (
    common, default_forecasts)


DEFAULT_SITEFILE = resource_filename(
    Requirement.parse('solarforecastarbiter'),
    'solarforecastarbiter/io/reference_observations/'
    'srml_reference_sites.json')


# maps the desired variable names to those returned by pvlib.iotools
srml_variable_map = {
    'ghi_': 'ghi',
    'dni_': 'dni',
    'dhi_': 'dhi',
    'wind_speed_': 'wind_speed',
    'temp_air_': 'air_temperature',
}


# maps SolarForecastArbiter interval_label to the SRML infix which
# designates the time resolution of each file. The list of file types
# is tried in order, so file types starting with 'P' designating
# processed data are listed first, such that if processed data exists
# we retrieve that first.
FILE_TYPE_MAP = {
    1: ['PO', 'RO'],
    5: ['PF', 'RF'],
    15: ['PQ', 'RQ'],
    60: ['PH', 'RH'],
}


logger = logging.getLogger('reference_data')


def adjust_site_parameters(site):
    """Inserts modeling parameters for sites with pv measurments

    Parameters
    ----------
    site: dict

    Returns
    -------
    dict
        Copy of inputs plus a new key 'modeling_parameters'.
    """
    return common.apply_json_site_parameters(DEFAULT_SITEFILE, site)


def request_data(site, year, month):
    """Makes a request for each file type until successful or we
    run out of filetypes.

    Parameters
    ----------
    site: :py:class:`solarforecastarbiter.datamodel.Site`
    year: int
        The year of the data to request.
    month: int
        The month of the data to request.

    Returns
    -------
    DataFrame
        A month of SRML data.
    """
    extra_params = common.decode_extra_parameters(site)
    station_code = extra_params['network_api_abbreviation']
    interval_length = extra_params['observation_interval_length']
    file_types = FILE_TYPE_MAP[interval_length]
    for file_type in file_types:
        # The list file_types are listed with processed data
        # file types first. On a successful retrieval we return
        # the month of data, otherwise we log info and continue
        # until we've exhausted the list.
        try:
            srml_month = iotools.read_srml_month_from_solardat(
                station_code, year, month, file_type)
        except error.URLError:
            logger.warning(f'Could not retrieve {file_type} for SRML data '
                           f'for site {site.name} on {year}/{month} .')
            logger.debug(f'Site abbreviation: {station_code}')
            continue
        except pd.errors.EmptyDataError:
            logger.warning(f'SRML returned an empty file for station '
                           f'{site.name} on {year}/{month}.')
            continue
        else:
            return srml_month
    logger.warning(f'Could not retrieve data for site {site.name} on '
                   f'{year}/{month}.')


def fetch(api, site, start, end):
    """Retrieve observation data for a srml site between start and end.

    Parameters
    ----------
    api : :py:class:`solarforecastarbiter.io.api.APISession`
        An APISession with a valid JWT for accessing the Reference Data
        user.
    site : :py:class:`solarforecastarbiter.datamodel.Site`
        Site object with the appropriate metadata.
    start : datetime
        The beginning of the period to request data for. Must include timezone.
    end : datetime
        The end of the period to request data for. Must include timezone.

    Returns
    -------
    data : pandas.DataFrame
        All of the requested data concatenated into a single DataFrame.

    Raises
    ------
    TypeError
        If start and end have different timezones, or if they do not include a
        timezone.
    """
    month_dfs = []
    start_year = start.year
    start_month = start.month
    # Retrieve each month file necessary
    if start.tzinfo != end.tzinfo:
        raise TypeError('start and end cannot have different timezones')
    while start_year * 100 + start_month <= end.year * 100 + end.month:
        logger.info(f'Requesting data for SRML site {site.name}'
                    f' for {start_year}-{start_month}')
        srml_month = request_data(site, start_year, start_month)
        if srml_month is not None:
            month_dfs.append(srml_month)
        start_month += 1
        if start_month > 12:
            start_month = 1
            start_year += 1
    try:
        all_period_data = pd.concat(month_dfs)
    except ValueError:
        logger.warning(f'No data available for site {site.name} '
                       f'from {start} to {end}.')
        return pd.DataFrame()
    var_columns = [col for col in all_period_data.columns
                   if '_flag' not in col]
    power_columns = [col for col in var_columns
                     if col.startswith('5')]
    # adjust power from watts to megawatts
    for column in power_columns:
        all_period_data[column] = all_period_data[column] / 1000000
    all_period_data = all_period_data.loc[start:end, var_columns]

    # remove possible trailing NaNs, it is necessary to do this after slicing
    # because SRML data has nighttime data prefilled with 0s through the end of
    # the month. This may not be effective if a given site has more than a 24
    # hour lag, which will cause last_valid_index to return the latest
    # timestamp just before sunrise, but will suffice for the typical lag on
    # the order of hours.
    all_period_data = all_period_data[:all_period_data.last_valid_index()]

    return all_period_data


def initialize_site_observations(api, site):
    """Creates an observation at the site for each variable in
    an SRML site's file.

    Parameters
    ----------
    api: :py:class:`solarforecastarbiter.io.api.APISession`

    site : :py:class:`solarforecastarbiter.datamodel.Site
        The site object for which to create Observations.

    Notes
    -----
    Since variables are labelled with an integer instrument
    number, Observations are named with their variable and
    instrument number found in the source files.

    e.g. A SRML file contains two columns labelled, 1001, and
    1002. These columns represent GHI at instrument 1 and
    instrument 2 respectively. The `pvlib.iotools` package
    converts these to 'ghi_1' and 'ghi_2' for us. We use these
    labels to differentiate between measurements recorded by
    different instruments.
    """
    # Request ~month old data at initialization to ensure we get a response.
    start = pd.Timestamp.utcnow() - pd.Timedelta('30 days')
    end = start
    try:
        extra_params = common.decode_extra_parameters(site)
    except ValueError:
        logger.warning('Cannot create reference observations at MIDC site '
                       f'{site.name}, missing required parameters.')
        return
    # use site name without network here to build
    # a name with the original column label rather than
    # the SFA variable
    site_name = common.site_name_no_network(site)
    try:
        site_df = fetch(api, site, start, end)
    except error.HTTPError:
        logger.error('Could not find data to create observations '
                     f'for SRML site {site_name}.')
        return
    else:
        if site_df is None:
            logger.error('Could not find data to create observations '
                         f'for SRML site {site_name}.')
            return
        for variable in srml_variable_map.keys():
            matches = [col for col in site_df.columns
                       if col.startswith(variable)]
            for match in matches:
                observation_extra_parameters = extra_params.copy()
                observation_extra_parameters.update({
                    'network_data_label': match})
                try:
                    # Here, we pass a name with match instead of variable
                    # to differentiate between multiple observations of
                    # the same variable
                    common.create_observation(
                        api, site, srml_variable_map[variable],
                        name=f'{site_name} {match}',
                        interval_label='beginning',
                        extra_params=observation_extra_parameters)
                except HTTPError as e:
                    logger.error(
                        f'Failed to create {variable} observation at Site '
                        f'{site.name}. Error: {e.response.text}')
        with open(DEFAULT_SITEFILE) as fp:
            obs_metadata = json.load(fp)['observations']
        for obs in obs_metadata:
            obs_site_extra_params = json.loads(obs['site']['extra_parameters'])
            if obs_site_extra_params['network_api_id'] == extra_params[
                    'network_api_id']:
                obs['site'] = site
                observation = Observation.from_dict(obs)
                common.check_and_post_observation(api, observation)


def initialize_site_forecasts(api, site):
    """
    Create a forecasts for each variable measured at the site

    Parameters
    ----------
    api : :py:class:`solarforecastarbiter.io.api.APISession`
        An active Reference user session.
    site : :py:class:`solarforecastarbiter.datamodel.Site`
        The site object for which to create Forecasts.
    """
    variables = list(srml_variable_map.values())
    if isinstance(site, SolarPowerPlant):
        variables += ['ac_power', 'dc_power']
    common.create_forecasts(
        api, site, variables,
        default_forecasts.TEMPLATE_FORECASTS)


def update_observation_data(api, sites, observations, start, end):
    """Post new observation data to a list of SRML Observations
    from start to end.

    api : :py:class:`solarforecastarbiter.io.api.APISession`
        An active Reference user session.
    sites: list of :py:class:`solarforecastarbiter.datamodel.Site`
        List of all reference sites as Objects
    observations: list of :py:class:`solarforecastarbiter.datamodel.Observation`
        List of all reference observations as Objects
    start : datetime
        The beginning of the period to request data for.
    end : datetime
        The end of the period to request data for.
    """  # noqa
    srml_sites = common.filter_by_networks(sites, 'UO SRML')
    for site in srml_sites:
        common.update_site_observations(api, fetch, site, observations,
                                        start, end)



import logging
from urllib.error import URLError


import pandas as pd
from pvlib import iotools

from solarforecastarbiter.io.reference_observations import (
    common, default_forecasts)


CRN_URL = 'https://www1.ncdc.noaa.gov/pub/data/uscrn/products/subhourly01/'

crn_variables = ['ghi', 'air_temperature', 'relative_humidity', 'wind_speed']

logger = logging.getLogger('reference_data')


def get_filename(site, year):
    """Get the applicable file name for CRN a site on a given date.
    """
    extra_params = common.decode_extra_parameters(site)
    network_api_id = extra_params['network_api_id']
    filename = f'{year}/CRNS0101-05-{year}-{network_api_id}.txt'
    return CRN_URL + filename


def fetch(api, site, start, end):
    """Requests data for a CRN site containing the requested start,
    end interval.

    Parameters
    ----------
    api : io.APISession
        An APISession with a valid JWT for accessing the Reference Data
        user.
    site : datamodel.Site
        Site object with the appropriate metadata.
    start : datetime
        The beginning of the period to request data for.
    end : datetime
        The end of the period to request data for.
    realtime : bool
        Whether or not to look for realtime data. Note that this data is
        raw, unverified data from the instruments.

    Returns
    -------
    data : pandas.DataFrame
        All of the requested data concatenated into a single DataFrame.
    """
    # CRN uses yearly files. To ensure we request the correct years,
    # we make a  request for each year between start and end
    year_dfs = []
    for year in range(start.year, end.year + 1):
        filename = get_filename(site, year)
        logger.info(f'requesting data for {site.name} on {year}')
        logger.debug(f'CRN filename: {filename}')
        try:
            crn_year = iotools.read_crn(filename)
        except URLError:
            logger.warning(f'Could not retrieve CRN data for site '
                           f'{site.name} for year {year}.')
            logger.debug(f'Failed CRN URL: {filename}.')
        else:
            year_dfs.append(crn_year)
    try:
        all_period_data = pd.concat(year_dfs)
    except ValueError:
        logger.warning(f'No data available for site {site.name} '
                       f'from {start} to {end}.')
        return pd.DataFrame()
    all_period_data = all_period_data.rename(
        columns={'temp_air': 'air_temperature'})
    return all_period_data


def initialize_site_observations(api, site):
    """Create an observation for each available variable at the SOLRAD site.

    Parameters
    ----------
    api : solarforecastarbiter.io.api.APISession
        An active reference user session
    site : solarforecastarbiter.datamodel.Site

    """
    for variable in crn_variables:
        common.create_observation(api, site, variable)


def initialize_site_forecasts(api, site):
    """
    Create a forecasts for each variable measured at the site

    Parameters
    ----------
    api : solarforecastarbiter.io.api.APISession
        An active Reference user session.
    site : datamodel.Site
        The site object for which to create Forecasts.
    """
    common.create_forecasts(
        api, site, crn_variables,
        default_forecasts.TEMPLATE_DETERMINISTIC_NWP_FORECASTS)


def update_observation_data(api, sites, observations, start, end):
    """ Post new observation data to all reference observations at each
    USCRN site between start and end.

    api : solarforecastarbiter.io.api.APISession
        An active Reference user session.
    sites: list of solarforecastarbiter.datamodel.Site
        List of all reference sites
    observations: list of solarforecastarbiter.datamodel.Observation
        List of all reference observations.
    start : datetime
        The beginning of the period to request data for.
    end : datetime
        The end of the period to request data for.
    """
    crn_sites = common.filter_by_networks(sites, 'NOAA USCRN')
    for site in crn_sites:
        common.update_site_observations(
            api, fetch, site, observations, start, end)
